/* Linux console support for xz80, copyright (C) 1994/5 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Header file linuxcon.h - linux svgalib replacements for random stuff,
 * and some useful X data type -> useable alternative hacks, so we don't
 * even need X installed.
 */

#include "rawkey.h"	/* for keyboard defines */

#if SCALE > 1
#error "!! must have SCALE = 1 for linux svgalib support !!"
#endif


#if 1
#define MODE_TO_USE	G320x200x256	/* very quick */
#define MODE_XSIZE	320
#define MODE_YSIZE	200
/* offset on real screen of speccy 256x192 screen (rest is border) */
#define SPECSCRN_XPOS	32
#define SPECSCRN_YPOS	4
#else
#define MODE_TO_USE	G320x240x256	/* gives good aspect ratio */
#define MODE_XSIZE	320
#define MODE_YSIZE	240
#define SPECSCRN_XPOS	32
#define SPECSCRN_YPOS	24
#endif


/* we put our pointer to the real speccy 256x192 screen in 'image'.
 */

/* most of this stuff isn't too important. */
typedef int Display;
typedef int Colormap;
typedef int Window;
typedef int Screen;
typedef unsigned char XImage;
typedef int GC;

/* resources stuff */
typedef struct {
  char *option,*specifier;
  int argKind;
  char *value;
  } XrmOptionDescRec;

typedef int XrmDatabase;
typedef struct {
  char *addr;
  int size;
  } XrmValue;

#define XrmoptionNoArg		1
#define XrmoptionSepArg		2
#define XrmoptionIsArg		2	/* treat last two the same */


#define DefaultDepthOfScreen(x)		(8)

typedef struct {
  int red,green,blue;
  int flags;
  int pixel;
  } XColor;

typedef struct {
  int type;
  int key;
  } XEvent;

typedef XEvent XKeyEvent;
typedef int KeySym;

/* just so it'll compile really, not used */
#define WhitePixel(x,y)		15
#define BlackPixel(x,y)		0
#define DoRed			1
#define DoGreen			2
#define DoBlue			4

#define True			1
#define False			0

#define KeyPress		1
#define KeyRelease		2

#define XCloseDisplay(x)	linux_graphics_off()
#define XAutoRepeatOn(x)	0
#define XFlush(x)		0
#define XClearWindow(x,y)	linux_draw_border()
#define XrmMergeDatabases(x,y)	0
#define XSetWindowBackground(a,b,c) 0

/* some random key mangles */

#define XK_F10		0x44
#define XK_F11		0x57
#define XK_F12		0x58
#define XK_F1		0x3B
#define XK_F2		0x3C
#define XK_F3		0x3D
#define XK_F4		0x3E
#define XK_F5		0x3F
#define XK_F6		0x40
#define XK_F7		0x41
#define XK_F8		0x42
#define XK_F9		0x43

#define XK_Return	ENTER_KEY
#define XK_Control_L	LEFT_CTRL
#define XK_Control_R	0x10000001	/* just a big, out-of-the-way number */
#define XK_Shift_L	LEFT_SHIFT
#define XK_Shift_R	RIGHT_SHIFT
#define XK_Alt_L	LEFT_ALT
#define XK_Alt_R	0x10000002
#define XK_Meta_L	0x10000003
#define XK_Meta_R	0x10000004
#define XK_BackSpace	BACKSPACE
#define XK_Delete	DELETE_KEY
#define XK_Escape	ESCAPE_KEY
#define XK_Tab		TAB_KEY
#define XK_Up		CURSOR_UP
#define XK_Down		CURSOR_DOWN
#define XK_Left		CURSOR_LEFT
#define XK_Right	CURSOR_RIGHT
#define XK_minus	GRAY_MINUS
#define XK_plus		GRAY_PLUS

/* these ones need to be converted to XK_underscore, etc. in linux_get_keys()
 * as we can't use a function in the switch() in spectrum.c:check_x_events().
 * They are currently only translated correctly for a UK keyboard. :-(
 */
#define XK_underscore	0x11000001	/* shifted */
#define XK_equal	0x11000002
#define XK_semicolon	0x11000003
#define XK_colon	0x11000004	/* shifted */
#define XK_apostrophe	0x11000005
#define XK_comma	0x11000006
#define XK_period	0x11000007
#define XK_less		0x11000008	/* shifted */
#define XK_greater	0x11000009	/* shifted */
#define XK_slash	0x1100000a
#define XK_question	0x1100000b	/* shifted */
#define XK_backslash	0x1100000c
#define XK_bar		0x1100000d	/* shifted */

#define NoKeyPressed	0x10000000

#define CTRL_Z		0x1100000e

extern void linux_graphics_on(void);
extern void linux_graphics_off(void);
extern void linux_get_keys(XEvent *xev);
extern int XAllocColor(Display *,Colormap,XColor *);
extern void linux_refresh_screen(void);
extern void linux_temp_text(void);
extern void linux_text_end(void);
extern void XrmInitialize(void);
extern void XrmParseCommand(XrmDatabase *,XrmOptionDescRec *,int,char *,
							int *,char **);
extern int XrmGetResource(XrmDatabase,char *,char *,char **,XrmValue *);
extern void XShmPutImage();
extern void XPutImage();
