/* xz80 - an X11-based ZX Spectrum Emulator.
 * Copyright (C) 1994 Ian Collier.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<stdio.h>
#include"z80.h"
#include <sys/time.h>
#include <signal.h>
unsigned char mem[65536];
unsigned char *dfile=mem+0x4000;
unsigned char *attrfile=mem+0x5800;
unsigned char *memptr[4]={mem,mem+0x4000,mem+0x8000,mem+0xc000};
int memattr[4]={0,2,1,1};
char squares[768];
unsigned char snapa,snapf,snapb,snapc,snapd,snape,snaph,snapl,
      snapa1,snapf1,snapb1,snapc1,snapd1,snape1,snaph1,snapl1,
      snapi,snapr,snapiff1,snapiff2,snapim;
unsigned short snapix,snapiy,snapsp,snappc;
int screen_dirty;
int hsize=256*SCALE,vsize=192*SCALE;
int load_snap=0;
unsigned long frames=0;
static int interrupted=0;

void sighandler(sig, code, scp, addr)
int sig, code;
struct sigcontext *scp;
char *addr;
{
   interrupted=1;
}

void hacks();
void load();

#ifndef SA_RESTART
#define SA_RESTART 0
#endif

main(argc,argv)
int argc;
char **argv;
{
   long elapsed;
   double spectime;
   int i;
   static struct sigaction act;
   act.sa_handler=sighandler;
   sigemptyset(&act.sa_mask);
   act.sa_flags=SA_RESTART;
   screen_dirty=1;
   for(i=0;i<768;i++)squares[i]=1;
   startup(&argc,argv);
   if(argc>1){
      load_snap=1;
      strcpy(snapname,argv[1]);
   }
   load(mem);
   hacks();
   refresh();
   if(sigaction(SIGALRM,&act,(struct sigaction*)0)){
      perror("Couldn't set timer signal handler");
      closedown();
      exit(1);
   }
#if !defined(DEBUG)
#ifdef SUN_AUDIO
   if(soundfd<0)
#endif
   itimeron();
#endif
   startwatch(0);
   mainloop();
   elapsed=stopwatch();
   closedown();
   spectime=frames*(double)config.cycles/3.5e6;
   fprintf(stderr,"Spectrum time: %.2f\nReal time: %lu.%03lu\nRatio: %.3f\n",
      spectime,elapsed/1000,elapsed%1000,spectime/elapsed*1000.);
}

int refresh_ct=1;

int interrupt(){
   static int flash_ct=flash_speed;

   int ans=check_events(1);
   if(!--flash_ct && config.flashing){
      flash_ct=flash_speed;
      flash();
   }
   if(!--refresh_ct){
      refresh_ct=refresh_frames;
      refresh();
   }
   eachframe();
#if !defined(DEBUG)
#ifdef SUN_AUDIO
   if(soundfd<0)
#endif
   while(!interrupted)sigpause(0);
   interrupted=0;
#endif
   return ans;
}

void load(x)
unsigned char *x;
{
   int i;
   FILE *f;
   if(!findfile(config.rom,0))f=0;
   else f=fopen(config.rom,"r");
   if(!f)perror(config.rom);
   else fread(x,1,16384,f),fclose(f);
   srand(getpid());
   for(i=16384;i<65536;i++)x[i]=rand()/256;
}

void hacks(){
   if(config.input_hack){
      static unsigned char oldip[]={253,203,1,110,200,58,8,92,253,203,1,174};
      static unsigned char newip[]={33,59,92,118,203,110,200,58,8,92,203,174};
      if(memcmp(memptr[0]+0x10b0,oldip,12))
         fputs("Input hack failed\n",stderr);
      else memcpy(memptr[0]+0x10b0,newip,12);
   }
   if(config.load){
      if(memptr[0][0x0556]!=0x14 || memptr[0][0x0557]!=8)
         fputs("Loader hack failed\n",stderr),
         config.load=0;
      else memptr[0][0x0556]=0xed,memptr[0][0x0557]=0xfc;
   }
   if(config.save){
      if(memptr[0][0x04c2]!=0x21 || memptr[0][0x04c3]!=0x3f)
         fputs("Saver hack failed\n",stderr),
         config.save=0;
      else memptr[0][0x04c2]=0xed,memptr[0][0x04c3]=0xfd;
   }
   if(config.noprompt){
      static unsigned char old[]={205,10,12,253,203,2,238,205,212,21,
                     221,229,17,17,0,175,205,194,4,221,225,6,50,118};
      static unsigned char new[]={0,0,0,0,0,0,0,0,0,0,
                     221,229,17,17,0,175,205,194,4,221,225,0,0,0};
      if(memcmp(memptr[0]+0x097a,old,24))
         fputs("Save prompt hack failed\n",stderr);
      else memcpy(memptr[0]+0x097a,new,24);
   }
}

