/* Utility routines for xz80, copyright (C) 1994 Ian Collier.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef sgi
#define _BSD_TIME
#endif

#include<stdio.h>
#include"z80.h"
#include<dirent.h>
#include<memory.h>
#include<string.h>
#include<errno.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/time.h>
#include<sys/param.h>
#include<unistd.h>

#ifdef LINUX_SVGALIB
#include "linuxcon.h"
#endif

static struct itimerval ival={{0,0},{0,0}};
static struct timeval watch;
static unsigned long millisec;
static int stopped;

char snapname[MAXPATHLEN];
char snapsavename[MAXPATHLEN];

void itimeron(){
   ival.it_interval.tv_usec=config.usec-200;  /* allow 200 usec overhead */
   ival.it_value.tv_usec=config.usec-200;
   if(setitimer(ITIMER_REAL,&ival,(struct itimerval*)0)){
      perror("Couldn't start interval timer");
      closedown();
      exit(1);
   }
}

void itimeroff(){
   ival.it_interval.tv_usec=0;
   ival.it_value.tv_usec=0;
   if(setitimer(ITIMER_REAL,&ival,(struct itimerval*)0))
      perror("Couldn't stop interval timer");
}

void startwatch(flag)
int flag;           /* 0 -> reset stopwatch; 1 -> carry on */
{
   if(!flag)millisec=0;
   gettimeofday(&watch,(struct timezone *)0);
   stopped=0;
}

unsigned long stopwatch(){
   unsigned long oldtime;
   unsigned long newtime;
   if(stopped)return millisec;
   oldtime=watch.tv_sec*1000+watch.tv_usec/1000;
   gettimeofday(&watch,(struct timezone *)0);
   newtime=watch.tv_sec*1000+watch.tv_usec/1000;
   stopped=1;
   return millisec+=(newtime-oldtime);
}

#if defined(__hpux) || ((defined(__SVR4)||defined(__svr4__)) && defined(sun))/* solaris */
#define getwd(buf) getcwd(buf,sizeof buf)
#else
char *getwd();
#endif

int findfile(name,types)/* look in LIB path for name; copy result back */
char *name;         /* into given pointer (hope it's long enough!) */
int types;          /* search order: name; name.zxf; name.sna; name.z80. */
                    /* types=0 -> search for name;                 */
                    /* types=1 -> search for name name.tap (in that order) */
                    /* types=3 -> search for name{,.sna,.z80}      */
                    /* types=4 -> search for all the above         */
                    /* returns 0 (+errno) if not found, 1 if found */
{
   char *gn;        /* the given name */
   char *path;      /* an element of the path */
   char *pathend;   /* end of this element */
   char *basename;  /* basename of the file to search for */
   int baselen;     /* length of basename */
   int pathlen;     /* length of path */
   int gpathlen;    /* length of path component in given name */
   int tmplen;      /* length of temporary filename buffer */
   static char tmp[MAXPATHLEN];  /* temporary filename buffer */
   static char ans[MAXPATHLEN];  /* anything found so far */
   DIR *dp;
   struct dirent *dir;
   int found=10;    /* position in search order above or a high number */
   int copy;
   extern int errno;

/* split name into pathname and basename */
   if((basename=strrchr(name,'/')))gpathlen=basename++-name;
   else gpathlen=0,basename=name;
   baselen=strlen(basename);
/* find out where to look. */
   gn=name;
   if(gn[0]=='.' && gn[1]=='/'){      /* special case for files in ./ */
      path=".";                       /* make path="." and remove the */
      gn+=2;                          /* "." from the name.  This     */
      gpathlen-=2;                    /* causes "." to be expanded.   */
   }
   else if(gn[0]!='/' && gpathlen)    /* for relative path names,     */
      path=".";                       /* prepend current dir name     */
   else if(gn[0]=='/'){
      path="";                        /* path given; prepend nothing  */
      if(types==0)
         return !access(gn,0); /* don't search if the full name given */
   }
   else path=config.lib;              /* usually search lib path */
/* scan each directory in the path */
   while(found!=1 && path){
      if((pathend=strchr(path,':')))pathlen=pathend++-path;
      else pathlen=strlen(path);
      if(pathlen==1 && path[0]=='.'){ /* expand "." */
         if(!getwd(tmp))strcpy(tmp,".");
         tmplen=strlen(tmp);
      }
      else memcpy(tmp,path,tmplen=pathlen);
      if(gpathlen && gn[0]!='/' && pathlen)tmp[tmplen++]='/';
      memcpy(tmp+tmplen,gn,gpathlen);
      tmp[tmplen+=gpathlen]=0;
      if((dp=opendir(tmp))){
/* examine each name in the directory */
         while(found!=1 && (dir=readdir(dp))){
            if(memcmp(dir->d_name,basename,baselen))
               continue;        /* check that it starts with basename */
            copy=0;
            if(!dir->d_name[baselen])copy=found=1;
            else if(types==1 && found>2 && !strcasecmp(dir->d_name+baselen,".tap"))
               copy=found=2;
            else if(types>3 && found>2 && !strcasecmp(dir->d_name+baselen,".zxf"))
               copy=found=2;
            else if(types>1 && found>3 && !strcasecmp(dir->d_name+baselen,".sna"))
               copy=found=3;
            else if(types>2 && found>4 && !strcasecmp(dir->d_name+baselen,".z80"))
               copy=found=4;
            if(copy)
               strcpy(ans,tmp),
               ans[tmplen]='/',
               strcpy(ans+tmplen+1,dir->d_name);
         }
         closedir(dp);
      }
      path=pathend;
   }
   if(found>5){
      errno=ENOENT;
      return 0;
   }
   strcpy(name,ans);
   return 1;
}

int snapload(){   /* try to load snapname and return 1 if successful */
   FILE *fp;
   struct stat st;
   int sna=1;
   int l;
   /* try to open it */
   if(!findfile(snapname,3) || !(fp=fopen(snapname,"r"))){
      perror(snapname);
      return 0;
   }
   /* try to decide what it is */
   if(fstat(fileno(fp),&st) || !S_ISREG(st.st_mode)){
      l=strlen(snapname);
      if(l>4 && (
            !strcmp(snapname+l-4,".z80")
         || !strcmp(snapname+l-4,".Z80")
        )) sna=0;
   }
   else if(st.st_size!=49179)sna=0;
   /* any way to detect a .Z80 file reliably? */

   if(sna)l=sna_load(fp);
   else l=z80_load(fp);
   fclose(fp);
   return l;
}

static unsigned short get2(fp)
FILE *fp;
{
   unsigned char l=getc(fp);
   unsigned char h=getc(fp);
   return (h<<8)|l;
}

static void put2(w,fp)
unsigned short w;
FILE *fp;
{
   putc(w&255,fp);
   putc(w>>8,fp);
}

int sna_load(fp)
FILE *fp;
{
   int i;
   snapi=getc(fp);
   snapl1=getc(fp);
   snaph1=getc(fp);
   snape1=getc(fp);
   snapd1=getc(fp);
   snapc1=getc(fp);
   snapb1=getc(fp);
   snapf1=getc(fp);
   snapa1=getc(fp);
   snapl=getc(fp);
   snaph=getc(fp);
   snape=getc(fp);
   snapd=getc(fp);
   snapc=getc(fp);
   snapb=getc(fp);
   snapiy=get2(fp);
   snapix=get2(fp);
   snapiff2=(getc(fp)>>2)&1;
   snapiff1=snapiff2;
   snapr=getc(fp);
   snapf=getc(fp);
   snapa=getc(fp);
   snapsp=get2(fp);
   snapim=getc(fp);
   if(snapim>0)snapim++;
   out(0,0,254,(getc(fp)&7)|8);
   if(ferror(fp)){
      perror(snapname);
      return 0;
   }
   if(feof(fp)){
      fprintf(stderr,"Premature EOF on %s\n",snapname);
      return 0;
   }
   for(i=0;i<768;squares[i++]=1);
   screen_dirty++;
   for(i=1;i<4;i++)fread(memptr[i],1,16384,fp);
   if(ferror(fp))perror(snapname);
   else if(feof(fp))
      fprintf(stderr,"Warning: premature EOF on %s\n",snapname);
   else if(getc(fp)!=EOF)
      fprintf(stderr,"Warning: unread bytes in %s\n",snapname);
   snappc=fetch2(snapsp);
   snapsp+=2;
   return 1;
}

int z80_load(fp)
FILE *fp;
{
   int i;
   int ch;
   int z;
   int addr=16384;

   snapf=getc(fp);
   snapa=getc(fp);
   snapc=getc(fp);
   snapb=getc(fp);
   snapl=getc(fp);
   snaph=getc(fp);
   snappc=get2(fp);
   snapsp=get2(fp);
   snapi=getc(fp);
   snapr=getc(fp);
   if((ch=getc(fp))==255)ch=1;
   snapr|=(ch&1)<<7;
   out(0,0,254,(ch>>1)&7);
   z=ch&0x20;
   snape=getc(fp);
   snapd=getc(fp);
   snapc1=getc(fp);
   snapb1=getc(fp);
   snape1=getc(fp);
   snapd1=getc(fp);
   snapl1=getc(fp);
   snaph1=getc(fp);
   snapf1=getc(fp);
   snapa1=getc(fp);
   snapiy=get2(fp);
   snapix=get2(fp);
   snapiff1=getc(fp)?1:0;
   snapiff2=getc(fp)?1:0;
   snapim=getc(fp)&3;
   if(snapim>0)snapim++;
   if(!snappc){
      fputs("New format .z80 files not supported\n",stderr);
      return 0;
   }
   if(ferror(fp)){
      perror(snapname);
      return 0;
   }
   if(feof(fp)){
      fprintf(stderr,"Premature EOF on %s\n",snapname);
      return 0;
   }
   for(i=0;i<768;squares[i++]=1);
   screen_dirty++;
   if(z){
      while((ch=getc(fp))!=EOF && addr<65536)
         if(ch==0xed){
            ch=getc(fp);
            if(ch==0xed){
               i=getc(fp);
               if(i==EOF){
                  fprintf(stderr,"Warning: premature EOF on %s\n",snapname);
                  return 1;
               }
               ch=getc(fp);
               while(i-- && addr<65536){store(addr,ch); addr++;}
            } else {
               store(addr,0xed);
               addr++;
               if(ch!=EOF && addr<65536){store(addr,ch); addr++;}
            }
         }
         else {store(addr,ch); addr++;}
      if(addr<65536)fprintf(stderr,"Warning: premature EOF on %s\n",snapname);
      else{
         /* discard any 'EDED00'x at the end of the file */
         (ch=getc(fp))==0xed && (ch=getc(fp))==0xed &&
         (ch=getc(fp))==0 && (ch=getc(fp));
         if(ch!=EOF)
            fprintf(stderr,"Warning: unread bytes in %s\n",snapname);
      }
   } else {
      for(i=1;i<4;i++)fread(memptr[i],1,16384,fp);
      if(ferror(fp))perror(snapname);
      else if(feof(fp))
         fprintf(stderr,"Warning: premature EOF on %s\n",snapname);
      else if(getc(fp)!=EOF)
         fprintf(stderr,"Warning: unread bytes in %s\n",snapname);
   return 1;
   }
}

void snapsave(){
   FILE *fp=fopen(snapsavename,"w");
   int i;
   if(!fp){
      perror(snapsavename);
      return;
   }
   snapsp-=2;
   store2(snapsp,snappc);
   putc(snapi,fp);
   putc(snapl1,fp);
   putc(snaph1,fp);
   putc(snape1,fp);
   putc(snapd1,fp);
   putc(snapc1,fp);
   putc(snapb1,fp);
   putc(snapf1,fp);
   putc(snapa1,fp);
   putc(snapl,fp);
   putc(snaph,fp);
   putc(snape,fp);
   putc(snapd,fp);
   putc(snapc,fp);
   putc(snapb,fp);
   put2(snapiy,fp);
   put2(snapix,fp);
   putc(snapiff2 ? 255 : 0,fp);
   putc(snapr,fp);
   putc(snapf,fp);
   putc(snapa,fp);
   put2(snapsp,fp);
   if(snapim>1)snapim--;
   putc(snapim,fp);
   putc(bordercr,fp);
   for(i=1;i<4;i++)fwrite(memptr[i],1,16384,fp);
   if(ferror(fp))perror(snapsavename);
   fclose(fp);
}

void requester(prompt,name,len)
char *prompt;
char *name;
int len;
{
   int timeron=0;
   extern int errno;
#  ifdef LINUX_SVGALIB
      linux_temp_text();
#  endif
   stopwatch();
   putchar(7);
   fputs(prompt,stdout);
   putchar(':');
   putchar(' ');
   fflush(stdout);
   while(!fgets(name,len,stdin)){
      if(errno==EINTR){
         timeron=1;
         itimeroff();
      } else {
         perror("Couldn't read standard input");
         name[0]=0;
         break;
      }
   }
   len=strlen(name)-1;
   if(len>=0 && name[len]=='\n')name[len]=0;
   if(timeron)itimeron();
#  ifdef LINUX_SVGALIB
      linux_text_end();
#  endif
}

#undef hl
#undef de

void multiloader(hl,a)
unsigned short hl;
unsigned char a;
{
   char *filetype;
   FILE *fp;
   int c;
   if(!config.multiload)return;
   if(!snapname[0]){
      fputs("No previous snapshot name - multiload ignored\n",stderr);
      return;
   }
   stopwatch();
   
   filetype=strrchr(snapname,'.');
   if(!filetype || strlen(filetype)>5)filetype=snapname+strlen(snapname);
   sprintf(filetype,"%d.dat",a);
   fp=fopen(snapname,"r");
   if(!fp)perror(snapname);
   else {
      while(hl && (c=getc(fp))!=EOF){store(hl,c); hl++;}
      if(ferror(fp))perror(snapname);
      else if(getc(fp)!=EOF)
         fprintf(stderr,"warning: unread bytes in %s\n",snapname);
      fclose(fp);
   }
   filetype[0]='.';
   filetype[1]=0;
   startwatch(1);
}

static FILE *infp=0,*outfp=0;
static int hdframe=0;
static char reqname[MAXPATHLEN];
static char newname[MAXPATHLEN];
static char xzx_magic[4];
static struct header {
   unsigned char type;
   char name[10];
   unsigned char length[2];
   unsigned char start[2];
   unsigned char line[2];
}  prepared={3,"xxxxxxxxxx",{0,0},{0,128},{0,0}};
/* two header structures used by BASIC; one at IX and one at IX-17. */
static struct header hd1,hd2;

void get_header(h,addr)
struct header *h;
unsigned short addr;
{
   int i;
   h->type=fetch(addr);
   addr++;
   for(i=0;i<10;i++){
      h->name[i]=fetch(addr);
      addr++;
   }
   h->length[0]=fetch(addr);addr++;
   h->length[1]=fetch(addr);addr++;
   h->start[0]=fetch(addr);addr++;
   h->start[1]=fetch(addr);addr++;
   h->line[0]=fetch(addr);addr++;
   h->line[1]=fetch(addr);
}

void put_header(h,addr)
struct header *h;
unsigned short addr;
{
   int i;
   store(addr,h->type);
   addr++;
   for(i=0;i<10;i++){
      store(addr,h->name[i]);
      addr++;
   }
   store(addr,h->length[0]);addr++;
   store(addr,h->length[1]);addr++;
   store(addr,h->start[0]);addr++;
   store(addr,h->start[1]);addr++;
   store(addr,h->line[0]);addr++;
   store(addr,h->line[1]);
}

/* File loader.  Emulates call to 0556.  Pseudo-op ED FC. */
int loader(pc,ix,de,a,c)
unsigned short pc,ix,de;
unsigned char a,c;
{
   if(pc>=16384 || !config.load)return -1;
   refresh();
   if(config.tap[0]){ /* use the given TAP file */
      int len,sum,ch,remain,ans;
      if(!tapefp && (taperead=1) &&
            (!findfile(config.tap,1) || !(tapefp=fopen(config.tap,"rb")))){
         perror(config.tap);
         fputs("Error opening tape file for input\n",stderr);
         return 1;
      }
      len=getc(tapefp);ch=getc(tapefp);
      if(ch==EOF){
         if(ferror(tapefp))perror("Loading error");
         else fputs("Loading error: End of file\n",stderr);
         return 1;
      }
      len+=256*ch;
      sum=ch=getc(tapefp);
      remain=len-1;
      ans=0;
      if(ch!=a)ans=2;
      else{
         if(de+2!=len)ans=2;
         if(de>len-2)de=len-2;
         while(de--){
            ch=getc(tapefp);
            remain--;
            if(ch==EOF){
               if(ferror(tapefp))perror("Loading error");
               else fputs("Loading error: End of file\n",stderr);
               ans=2;
               remain=0;
               break;
            }
            else if(c)store(ix,ch);
            else if((unsigned char)ch!=fetch(ix)){
               ans=2;
               break;
            }
            ix++;
            sum^=ch;
         }
         if(!ans){
            ch=getc(tapefp);
            remain--;
            sum^=ch;
            if(sum)ans=2;
         }
      }
      if(ans && remain)fseek(tapefp,remain,1);
      return ans;
   }
   /* else use XZ80 tape files */
   if(!a){  /* header */
      int i,basic=(ix<fetch2(23653)),wasbasic=basic;
      if(basic){
         get_header(&hd1,ix-17);
         if(hd1.type>=4)basic=0; /* ignore basic header for LOAD"" */
      }
      memcpy(reqname,hd1.name,10);
      for (i=9;i>=0 && reqname[i]==' ';i--);
      reqname[i+1]=0;
      if(i<0 || reqname[0]==(char)255)
         basic=0; 
      if(infp){ 
         fclose(infp),infp=0;
         if(frames-hdframe<5){
            /* The header was probably rejected by BASIC.  Avoid getting
               into an infinite loop by requesting another file name. */
            requester("Header rejected - enter another file name",
               reqname,sizeof reqname);
            if(!reqname[0])return 1;  /* request was rejected */
         }
      }
      else if(!basic){
         requester("Enter file name for load",reqname,sizeof reqname);
         if(!reqname[0])return 1;
      }
      while(!findfile(reqname,4) || !(infp=fopen(reqname,"r"))){
         perror(reqname);
         requester("Enter another file name",reqname,sizeof reqname);
         if(!reqname[0])return 1;
      }
      hdframe=frames;
      if (fread(xzx_magic,1,sizeof xzx_magic,infp)<sizeof xzx_magic ||
          fread((char*)&hd2,1,17,infp)<17 ||
          memcmp(xzx_magic,XZX_MAGIC,sizeof xzx_magic)) {
         /* it isn't an xz* file.  Fill in the header */
         static struct stat st;
         rewind(infp);
         memcpy((char*)&hd2,(char*)&prepared,17);
         if(basic) memcpy(hd2.name,hd1.name,10);
         else {
            for(i=0;i<10 && (hd2.name[i]=reqname[i]);i++);
            while(i<10)hd2.name[i++]=' ';
         }
         if(fstat(fileno(infp),&st)){
            fputs("fstat: ",stderr);
            perror(reqname);
            st.st_size=0;
         }
         hd2.length[0]=st.st_size%256;
         hd2.length[1]=st.st_size/256;
         put_header(&hd2,ix);
         if(st.st_size==49179 && wasbasic && !hd1.type){
            /* load a snapshot but only if it was a straight LOAD "foo"
               command */
            strcpy(snapname,reqname);
            load_snap=1;
            return 1;
         }
         else if(st.st_size>49152)
            fprintf(stderr,"%s has length %ld!\n",reqname,(long)st.st_size);
      } else { /* It is an xz* file.  Kludge the file name. */
         if(basic)memcpy(hd2.name,hd1.name,10);
         else {
            for(i=0;i<10 && (hd2.name[i]=reqname[i]);i++);
            while(i<10)hd2.name[i++]=' ';
         }
         put_header(&hd2,ix);
      }
      return 0;
   } else { /* data */
      int ans=0;
      int ch;
      if(!infp){  /* headerless load */
         requester("Enter file name for headerless load",
            reqname,sizeof reqname);
         if(!reqname[0])return 1;
         infp=fopen(reqname,"r");
         if(!infp){
            perror(reqname);
            return 2;
         }
      }
      while(de--){
         ch=getc(infp);
         if(ch==EOF){
            ans=2;
            if(ferror(infp))perror("Loading error");
            else fputs("Loading error: premature EOF\n",stderr);
            break;
         }
         else if(c)store(ix,ch);
         else if((unsigned char)ch!=fetch(ix)){
            ans=2;
            break;
         }
         ix++;
      }
      fclose(infp);
      infp=0;
      return ans;
   }
}

/* File saver.  Emulates call to 04C2.. Pseudo-op ED FD. */
int saver(pc,ix,de,a)
unsigned short pc,ix,de;
unsigned char a;
{
   if(pc>=16384 || !config.save)return -1;
   refresh();
   if(config.tap[0]){ /* use the given TAP file */
      int ch,sum;
      if(!tapefp || taperead){
         if(tapefp)fclose(tapefp);
         taperead=0;
         if(!(tapefp=fopen(config.tap,"a+b"))){
            perror(config.tap);
            fputs("Error opening tape file for output\n",stderr);
            return 1;
         }
      }
      /* a seek is needed between a read and a write, so just in case...*/
      fseek(tapefp,0,2);
      sum=de+2;
      putc(sum&255,tapefp);
      putc(sum/256,tapefp);
      putc(sum=a,tapefp);
      while(de--){
         putc(ch=fetch(ix),tapefp);
         ix++;
         sum^=ch;
      }
      putc(sum,tapefp);
      fflush(tapefp);
      fseek(tapefp,0,0); /* in case the next operation is a read */
      return 0;
   }
   /* else use XZ80 tape files */
   if(!a){ /* header */
      int i;
      get_header(&hd1,ix);
      if(outfp)fclose(outfp);  /* second successive header.  Not much we can do...*/
      memcpy(reqname,hd1.name,10);
      for(i=9;i>0&&reqname[i]==' ';i--);
      reqname[i+1]=0;
      strcpy(newname,reqname);
      strcat(newname,".bak");
      rename(reqname,newname);
      outfp=fopen(reqname,"w");
      if(!outfp){
         perror(reqname);
         return 1;
      }
      fwrite(XZX_MAGIC,1,sizeof xzx_magic,outfp);
   }
   else if(!outfp){ /* headerless save */
      requester("Enter file name for headerless save",reqname,sizeof reqname);
      if(!reqname[0])return 1;
      outfp=fopen(reqname,"w");
      if(!outfp){
         perror(reqname);
         return 1;
      }
   }
   while(de--){
      putc(fetch(ix),outfp);
      ix++;
   }
   if(a)fclose(outfp),outfp=0;
   return 0;
}

#define bool(x) (x?"on":"off")

void usage(){
   printf("This is xz80 version %s, copyright (C) 1994 Ian Collier\n",VERSION);
   puts("This is free software, and you are welcome to redistribute it");
   puts("under certain conditions.  xz80 comes with no warranty.");
   puts("For details, see the file COPYING.");
   putchar('\n');
   puts("Usage: xz80 [options] [snapshot-file]");
   puts("Available options:");
   printf(" -border <bool>      Enable the border [%s]\n",config.border==2?"full":bool(config.border));
   printf(" -cycles <num>       Specify number of T-states per frame [%d]\n",config.cycles);
   printf(" -display <display>  Specify X display\n");
   printf(" -flashing <bool>    Enable flashing attributes [%s]\n",bool(config.flashing));
   puts(  " -help or -?         Display this message");
   printf(" -inphack <bool>     Enable low-cpu idling [%s]\n",bool(config.input_hack));
   printf(" -kempstright <key>  Define key for kempston right [%s]\n",kempinit[0].keysym);
   printf(" -kempstleft  <key>  Define key for kempston left [%s]\n",kempinit[1].keysym);
   printf(" -kempstdown  <key>  Define key for kempston down [%s]\n",kempinit[2].keysym);
   printf(" -kempstup    <key>  Define key for kempston up [%s]\n",kempinit[3].keysym);
   printf(" -kempstfire  <key>  Define key for kempston fire [%s]\n",kempinit[4].keysym);
   printf(" -lib <path>         Specify library search path [%s]\n",config.lib);
   printf(" -loader <bool>      Enable tape-loader emulation [%s]\n",bool(config.load));
#ifdef DEBUG
   printf(" -log <file>         Specify log file [%s]\n",config.log);
#endif
#ifdef MITSHM
   printf(" -mitshm <bool>      Enable MIT-SHM [%s]\n",bool(config.mitshm));
#endif
   printf(" -mono               Use only black and white [%s]\n",bool(config.mono));
   printf(" -ms <int>           Specify number of milliseconds per frame [%d]\n",config.usec/1000);
   printf(" -multiload <bool>   Enable the multiload pseudo-op [%s]\n",bool(config.multiload));
   printf(" -noprompt <bool>    Enable immediate saving [%s]\n",bool(config.noprompt));
   printf(" -printer <bool>     Enable ZX Printer emulation [%s]\n",bool(config.printer));
   printf(" -printfile <file>   Specify the ZX Printer output file [%s]\n",config.print[0]?config.print:"none");
   printf(" -private            Use a private colormap [%s]\n",bool(config.private));
   printf(" -rom <file>         Specify the file to use as ROM image [%s]\n",config.rom);
   printf(" -rr <int>           Refresh screen every <int> frames\n");
#ifdef MITSHM
   printf(" -rrnoshm <int>      Like -rr but only works with MITSHM off [%d]\n",config.rrnoshm);
   printf(" -rrshm <int>        Like -rr but only works with MITSHM on [%d]\n",config.rrshm);
#else
   printf(" -rrnoshm <int>      Refresh screen every <int> frames [%d]\n",config.rrnoshm);
#endif
   printf(" -saver <bool>       Enable tape-saver emulation [%s]\n",bool(config.save));
   printf(" -tape <file>        Specify a .TAP-format tape file to use [%s]\n",config.tap[0]?config.tap:"none");
   puts(  " -version            Display version number and compile-time options");
#ifdef DEV_AUDIO
   printf(" -audiodev <file>    Specify the audio device file name [%s]\n",config.audio);
#ifdef DEV_AUDIO_IN
   printf(" -audioindev <file>  Specify the audio input device file name [%s]\n",config.audioin);
#endif
   printf(" -sample <int>       Specify the audio device sample rate [%d]\n",config.sample);
   printf(" -sound <bool>       Enable sound [%s]\n",bool(config.sound));
   printf(" -volume <int>       Specify the volume (0-256) [%d]\n",config.volume);
#endif
}
#undef bool

void version() {
   printf("This is xz80 version %s, copyright (C) 1994 Ian Collier\n",VERSION);
   puts("This is free software, and you are welcome to redistribute it");
   puts("under certain conditions.  xz80 comes with no warranty.");
   puts("For details, see the file COPYING.");
   putchar('\n');
   puts("Compile-time options:");
#ifdef MITSHM
   puts(" * MIT-SHM image transfer code is compiled in.");
#endif
#ifdef FANCYBORDER
   puts(" * The -border full option is implemented.");
#endif
#ifdef DEV_AUDIO
   puts(" * Audio emulation code is compiled in.");
#ifdef DEV_AUDIO_IN
   puts(" * Audio input emulation code is compiled in.");
#endif
#ifdef SUN_AUDIO
   puts(" * Audio is synchronised with the AUDIO_DRAIN ioctl.");
#endif
#endif
#ifdef LINUX_SVGALIB
   puts(" * The Linux svgalib is used for the display.");
#else
   printf(" * X11 pixel to Spectrum pixel scaling factor is %d:1.\n",SCALE);
#endif
   putchar('\n');
   puts("To see the default values for command-line options, please type xz80 -help");
}
