/* Utility routines for xz80, copyright (C) 1994 Ian Collier.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This file contains all the X-related stuff and all the spectrum
   hardware emulation.  Note that, although I did not actually load
   any part of xzx into my editor, several of the following routines
   were modelled on those of xzx and may bear very close resemblence.
   xzx was written by Des Herriott <dnh@mfltd.co.uk>, who has kindly
   given permission for me to include these routines in xz80.
   Modifications to these routines for the Linux console using svgalib
   were supplied by Russell Marks <mr216@greenwich.ac.uk> (together
   with the source files linuxcon.c, linuxcon.h and rawkeys.h).  */

#ifndef LINUX_SVGALIB
# include <X11/Xlib.h>
# include <X11/Xutil.h>
# include <X11/keysym.h>
# include <X11/Intrinsic.h>

# ifdef MITSHM
#  include <sys/ipc.h>
#  include <sys/shm.h>
#  include <X11/extensions/XShm.h>
# endif
#endif /* LINUX_SVGALIB */

#ifdef HAS_UNAME
#include <sys/utsname.h>
#endif

#ifdef SunKludge
char *shmat();
char *getenv();
#endif

#ifdef SUN_AUDIO
#include<sys/ioccom.h>
#include<sun/audioio.h>
#endif
#ifdef DEV_AUDIO
# include<fcntl.h>
# ifdef linux
#  include <sys/soundcard.h>
# endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"

#ifdef LINUX_SVGALIB
#include "linuxcon.h"
#endif

#define MAX_DISP_LEN 256
#define MAX_CHARS 80
#ifndef LINUX_FRAG
#define AUDIOBUFLEN 512
#else
#define AUDIOBUFLEN 1024
#endif

#include "xz80.icon"

struct _config config;
int refresh_screen;
int refresh_frames;
int bordercr=7;
int borderchange=0;
int soundfd=-1;
int soundinfd=-1;
FILE *tapefp=0;
int taperead=1;
struct _kempinit kempinit[]={
   /* table of resource names and default key names for kempston */
   { "kempstright", "p" },   /* bit 0 of port */
   { "kempstleft",  "o" },   /* bit 1 of port */
   { "kempstdown",  "a" },   /* bit 2 of port */
   { "kempstup",    "q" },   /* bit 3 of port */
   { "kempstfire",  "space" }/* bit 4 of port */
};
static KeySym kemptable[5];
unsigned char keyports[9]={255,255,255,255,255,255,255,255,255};
unsigned char kemport=0;
static int symbolshift,capsshift,unshift;
static struct {unsigned char port,mask;} keytable[]={
/*  SP      !        "        #        $        %        &        '  */
{7,0xfe},{3,0xfe},{3,0xfd},{3,0xfb},{3,0xf7},{3,0xef},{4,0xf7},{3,0xfd},
/*  (       )        *        +        ,        -        .        /  */
{4,0xfd},{4,0xfe},{4,0xfb},{6,0xfb},{7,0xf7},{6,0xf7},{8,0xfb},{0,0xef},
/*  0       1        2        3        4        5        6        7  */
{4,0xfe},{3,0xfe},{3,0xfd},{3,0xfb},{3,0xf7},{3,0xef},{4,0xef},{4,0xf7},
/*  8       9        :        ;        <        =        >        ?  */
{4,0xfb},{4,0xfd},{0,0xfd},{5,0xfd},{7,0xf7},{6,0xfb},{8,0xfb},{0,0xef},
/*  @       A        B        C        D        E        F        G  */
{3,0xfd},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
/*  H       I        J        K        L        M        N        O  */
{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
/*  P       Q        R        S        T        U        V        W  */
{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
/*  X       Y        Z        [        \        ]        ^        _  */
{0,0xfb},{5,0xef},{0,0xfd},{8,0xff},{8,0xff},{8,0xff},{4,0xef},{6,0xf7},
/*  `       a        b        c        d        e        f        g  */
{8,0xff},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
/*  h       i        j        k        l        m        n        o  */
{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
/*  p       q        r        s        t        u        v        w  */
{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
/*  x       y        z        {        |        }        ~       DEL */
{0,0xfb},{5,0xef},{0,0xfd},{8,0xff},{8,0xff},{8,0xff},{8,0xff},{8,0xff}
};

static unsigned short rgbvals[16][3]={
        { 0x0000, 0x0000, 0x0000}, { 0x0000, 0x0000, 0xbfff},
        { 0xbfff, 0x0000, 0x0000}, { 0xbfff, 0x0000, 0xbfff},
        { 0x0000, 0xbfff, 0x0000}, { 0x0000, 0xbfff, 0xbfff},
        { 0xbfff, 0xbfff, 0x0000}, { 0xbfff, 0xbfff, 0xbfff},
        { 0x0000, 0x0000, 0x0000}, { 0x0000, 0x0000, 0xffff},
        { 0xffff, 0x0000, 0x0000}, { 0xffff, 0x0000, 0xffff},
        { 0x0000, 0xffff, 0x0000}, { 0x0000, 0xffff, 0xffff},
        { 0xffff, 0xffff, 0x0000}, { 0xffff, 0xffff, 0xffff}
};

static int do_reverse=0;
static unsigned char reverse[256]; /* to hold table of reversed bytes,
                                 e.g. 00110100B -> 00101100B */
#if SCALE>1
#define shade(val,inkpap,paper)  (((val)&(inkpap))^(paper))
static unsigned long scaleup[256]; /* to hold table of scaled up bytes,
                                 e.g. 00110100B -> 0000111100110000B */
static unsigned char halftone[16][4]={    /* 8-by-4 bitmaps for halftoning */
        {0xff,0xff,0xff,0xff},   /* black */
        {0xee,0xff,0xbb,0xff},   /* blue */
        {0x55,0xff,0xaa,0xff},   /* red */
        {0x55,0xbb,0x55,0xff},   /* magenta */
        {0xaa,0x55,0xaa,0x55},   /* green */
        {0x51,0xaa,0x15,0xaa},   /* cyan */
        {0xaa,0x44,0xaa,0x00},   /* yellow */
        {0x11,0x00,0x44,0x00},   /* white */
        {0xff,0xff,0xff,0xff},   /* black */
        {0x77,0xdf,0x77,0xfd},   /* bright blue */
        {0xaa,0xdd,0xaa,0x77},   /* bright red */
        {0xae,0x55,0xea,0x55},   /* bright magenta */
        {0x55,0x22,0x55,0x88},   /* bright green */
        {0x88,0x22,0x88,0x22},   /* bright cyan */
        {0x00,0x40,0x00,0x04},   /* bright yellow */
        {0x00,0x00,0x00,0x00}    /* bright white */
};
#endif

static Display *display;
static Screen *scrptr;
static int screen;
static Window root;
static Colormap cmap;
static GC maingc;
static GC fggc, bggc;
static Window borderwin, mainwin;
static XImage *ximage;
       unsigned char *image;   /* not static because used by linuxcon.c */
#if defined(MITSHM) && !defined(LINUX_SVGALIB)
static XShmSegmentInfo xshminfo;
#endif
#ifdef FANCYBORDER
static int frame_sync=K48_SYNC;
static int linecycles=K48_LINE;
static XImage *bordximage;
       unsigned char *bordimage;
#if defined(MITSHM) && !defined(LINUX_SVGALIB)
static XShmSegmentInfo bordxshminfo;
#endif
#endif
static int linelen;
static int black,white;
static unsigned long pixels[16];
static int bytecycles;
static int framebytes;
static int invert=0;

/* Things starting with zxp are for the ZX Printer. */
static void zxpopen(),zxpclose(),zxpout();
static int zxpframes,zxpcycles,zxpspeed,zxpnewspeed,zxpheight,zxppixel,
           zxpstylus;
static FILE *zxpfile;
static char zxpline[256];

static XrmOptionDescRec opts[] = {
        {"-rom",        ".rom",        XrmoptionSepArg,NULL},
        {"-lib",        ".lib",        XrmoptionSepArg,NULL},
        {"-mitshm",     ".mitshm",     XrmoptionSepArg,NULL},
        {"-mono",       ".mono",       XrmoptionIsArg,NULL},
        {"-private",    ".private",    XrmoptionIsArg,NULL},
        {"-flashing",   ".flashing",   XrmoptionSepArg,NULL},
        {"-multiload",  ".multiload",  XrmoptionSepArg,NULL},
        {"-saver",      ".saver",      XrmoptionSepArg,NULL},
        {"-loader",     ".loader",     XrmoptionSepArg,NULL},
        {"-noprompt",   ".noprompt",   XrmoptionSepArg,NULL},
        {"-border",     ".border",     XrmoptionSepArg,NULL},
#ifdef DEV_AUDIO
        {"-sound",      ".sound",      XrmoptionSepArg,NULL},
        {"-sample",     ".sample",     XrmoptionSepArg,NULL},
        {"-audiodev",   ".audiodev",   XrmoptionSepArg,NULL},
#ifdef DEV_AUDIO_IN
        {"-audioindev", ".audioindev", XrmoptionSepArg,NULL},
#endif
        {"-volume",     ".volume",     XrmoptionSepArg,NULL},
#endif
        {"-kempstright",".kempstright",XrmoptionSepArg,NULL},
        {"-kempstleft", ".kempstleft", XrmoptionSepArg,NULL},
        {"-kempstup",   ".kempstup",   XrmoptionSepArg,NULL},
        {"-kempstdown", ".kempstdown", XrmoptionSepArg,NULL},
        {"-kempstfire", ".kempstfire", XrmoptionSepArg,NULL},

        {"-ms",         ".ms",         XrmoptionSepArg,NULL},
        {"-cycles",     ".cycles",     XrmoptionSepArg,NULL},
        {"-inphack",    ".inphack",    XrmoptionSepArg,NULL},
        {"-rrshm",      ".rrshm",      XrmoptionSepArg,NULL},
        {"-rrnoshm",    ".rrnoshm",    XrmoptionSepArg,NULL},
        {"-rr",         ".rr",         XrmoptionSepArg,NULL},
        {"-display",    ".display",    XrmoptionSepArg,NULL},
        {"-printfile",  ".printfile",  XrmoptionSepArg,NULL},
        {"-tape",       ".tape",       XrmoptionSepArg,NULL},
#ifdef DEBUG
        {"-log",        ".log",        XrmoptionSepArg,NULL},
#endif
        {"-printer",    ".printer",    XrmoptionSepArg,NULL}
};

static int is_local_server(name)
char *name;
{
#ifdef HAS_UNAME
   struct utsname un;
#else
   char sysname[MAX_DISP_LEN];
#endif

   if(name[0]==':')return 1;
   if(!strncmp(name,"unix",4))return 1;
   if(!strncmp(name,"localhost",9))return 1;

#ifdef HAS_UNAME
   uname(&un);
   if(!strncmp(name,un.sysname,strlen(un.sysname)))return 1;
   if(!strncmp(name,un.nodename,strlen(un.nodename)))return 1;
#else
   gethostname(sysname,MAX_DISP_LEN);
   if(!strncmp(name,sysname,strlen(sysname)))return 1;
#endif
   return 0;
}

static int bool(s)
char *s;
{
   return !(strcasecmp(s,"true") && strcasecmp(s,"on") &&
            strcasecmp(s,"yes"));
}

static void get_string(db,myname,myclass,resource,Default,result,size)
XrmDatabase db;
char *myname,*myclass,*resource,*Default,*result;
unsigned size;
{
   int i,len=0;
   char *ans;
   char name[80],class[80],*type[10];
   XrmValue rmvalue;
   sprintf(name,"%s.%s",myname,resource);
   sprintf(class,"%s.%c%s",myclass,
      isupper(*resource)?toupper(*resource):*resource,
      resource+1);
   if(XrmGetResource(db,name,class,type,&rmvalue)){
      if(!(ans=rmvalue.addr))ans="";
      else len=rmvalue.size;
      for (i=0;i<(sizeof opts)/(sizeof *opts);i++)
         if(opts[i].argKind==XrmoptionIsArg &&
               !(strcmp(ans,opts[i].option)&&strcmp(ans,opts[i].specifier))){
            ans="true",len=4;
            break;
         }
   }
   else if(Default)len=strlen(ans=Default);
   else ans="";
   while(len && isspace(*ans))len--,ans++;
   while(len && isspace(ans[len-1]))len--;
   if(len>size-1)len=size-1;
   memcpy(result,ans,len);
   result[len]=0;
}

static int get_bool(db,myname,myclass,resource,Default)
XrmDatabase db;
char *myname,*myclass,*resource;
int Default;
{
   char result[6];
   get_string(db,myname,myclass,resource,"",result,sizeof result);
   if(result[0])return bool(result);
   else return Default;
}

static int get_tri(db,myname,myclass,resource,Default)
XrmDatabase db;
char *myname,*myclass,*resource;
int Default;
{
   char result[6];
   get_string(db,myname,myclass,resource,"",result,sizeof result);
   if(!result[0])return Default;
   if(bool(result)) return 1;
   if(!strcasecmp(result,"full"))return 2;
   return 0;
}

static int get_int(db,myname,myclass,resource,Default)
XrmDatabase db;
char *myname,*myclass,*resource;
int Default;
{
   char result[25];
   int ans;
   get_string(db,myname,myclass,resource,"",result,sizeof result);
   if(sscanf(result,"%d",&ans)<1)return Default;
   else return ans;
}

static void get_defaults(display,dbptr,myclass)
Display *display;
XrmDatabase *dbptr;
char *myclass;
{
#ifndef LINUX_SVGALIB
   static char path[MAXPATHLEN];
   XrmDatabase dbtmp;
   char *dir=getenv("XAPPLRESDIR");
   char *s=XResourceManagerString(display);
   if(s)dbtmp=XrmGetStringDatabase(s);
   else{
      char *home=getenv("HOME");
      if(home){
         sprintf(path,"%s/.Xdefaults",home);
         dbtmp=XrmGetFileDatabase(path);
      }
      else dbtmp=0;
   }
#  ifdef XAPPLRESDIR
      if(!dir)dir=XAPPLRESDIR;
#  endif
   if(dir){
      sprintf(path,"%s/%s",dir,myclass);
      *dbptr=XrmGetFileDatabase(path);
   }
   else *dbptr=0;
   XrmMergeDatabases(dbtmp,dbptr);
#endif
}

static Display *open_display(argc,argv)
int *argc;
char **argv;
{
   int i;
   char *ptr;
   char *myname="xz80",*myclass="Xz80";
   char dispname[MAX_DISP_LEN];
   char keyname[KEYSYMLEN];
   XrmDatabase adb,db;
   Display *display;

   XrmInitialize();
   adb=0;
   XrmParseCommand(&adb,opts,(sizeof opts)/(sizeof *opts),
       myname,argc,argv);
#  ifdef LINUX_SVGALIB
      linux_graphics_on();
#  else
      get_string(adb,myname,myclass,"display","",dispname,MAX_DISP_LEN);
      if(!dispname[0])
         if((ptr=getenv("DISPLAY")))strcpy(dispname,ptr);
         else strcpy(dispname,":0.0");
      if(!(display=XOpenDisplay(dispname))){
         fprintf(stderr,"Unable to open display %s\n",dispname);
         exit(1);
      }
#  endif
   config.rrshm=config.rrnoshm=get_int(adb,myname,myclass,"rr",0);
   get_defaults(display,&db,myclass);
   XrmMergeDatabases(adb,&db);

   config.mitshm=get_bool(db,myname,myclass,"mitshm",1);
   config.mono=get_bool(db,myname,myclass,"mono",0);
   config.flashing=get_bool(db,myname,myclass,"flashing",1);
   config.sound=get_bool(db,myname,myclass,"sound",0);
   config.shaded=!config.mono;
   config.private=get_bool(db,myname,myclass,"private",0);
   config.input_hack=get_bool(db,myname,myclass,"inphack",1);
   config.printer=get_bool(db,myname,myclass,"printer",0);
   config.multiload=get_bool(db,myname,myclass,"multiload",1);
   config.load=get_bool(db,myname,myclass,"loader",1);
   config.save=get_bool(db,myname,myclass,"saver",1);
   config.noprompt=get_bool(db,myname,myclass,"noprompt",config.save);
   config.border=get_tri(db,myname,myclass,"border",1);

   if(!config.rrshm)config.rrshm=get_int(db,myname,myclass,"rrshm",2*SCALE);
   if(!config.rrnoshm)config.rrnoshm=get_int(db,myname,myclass,"rrnoshm",6*SCALE);
   config.usec=1000*get_int(db,myname,myclass,"ms",20);
   config.cycles=get_int(db,myname,myclass,"cycles",K48_CYCLES);
   if(config.cycles<=70000)frame_sync=K48_SYNC,linecycles=K48_LINE;
   else                    frame_sync=K128_SYNC,linecycles=K128_LINE;
#ifdef DEV_AUDIO
   config.sample=get_int(db,myname,myclass,"sample",8000);
   config.volume=get_int(db,myname,myclass,"volume",128);
   bytecycles=config.cycles*(1000000/config.usec)/config.sample;
   framebytes=config.cycles/bytecycles;
   if(framebytes+1>AUDIOBUFLEN){
      fprintf(stderr,"Audio buffer has size %d; needs %d bytes!\n",
         AUDIOBUFLEN,framebytes+1);
      framebytes=AUDIOBUFLEN-1;
      bytecycles=config.cycles/framebytes;
   }
#endif

   get_string(db,myname,myclass,"rom","spectrum.rom",config.rom,
              sizeof config.rom);
   get_string(db,myname,myclass,"lib",LIBPATH,config.lib,
              sizeof config.lib);
#ifdef DEBUG
   get_string(db,myname,myclass,"log","/tmp/xz80log",config.log,
              sizeof config.log);
#endif
   get_string(db,myname,myclass,"printfile",(char *)0,config.print,
              sizeof config.print);
   get_string(db,myname,myclass,"tape",(char *)0,config.tap,
              sizeof config.tap);
#ifndef LINUX_SVGALIB
   for(i=0;i<5;i++){ /* Get the kempston key symbols */
      get_string(db,myname,myclass,kempinit[i].resource,kempinit[i].keysym,
                keyname,KEYSYMLEN);
      kemptable[i]=XStringToKeysym(keyname);
      if(kemptable[i]==NoSymbol){
         fprintf(stderr,"Unknown keysym %s found; using %s\n",keyname,
                 kempinit[i].keysym);
         kemptable[i]=XStringToKeysym(kempinit[i].keysym);
      }
      else strcpy(kempinit[i].keysym,keyname);
   }
#endif
#ifdef DEV_AUDIO
   get_string(db,myname,myclass,"audiodev",AUDIODEV,config.audio,
              sizeof config.audio);
#ifdef DEV_AUDIO_IN
   get_string(db,myname,myclass,"audioindev",AUDIODEV,config.audioin,
              sizeof config.audioin);
#endif
#endif

   for (i=1;i<*argc;i++)
      if(argv[i][0]=='-'){
         if(!strcasecmp(argv[i],"-help") || argv[i][1]=='?')usage();
         else if(!strcmp(argv[i],"-version"))version();
         else fprintf(stderr,"Unrecognised option '%s'.  Type xz80 -help for a list of options.\n",argv[i]);
         XCloseDisplay(display);
         exit(1);
      }

   if(config.printer)zxpopen();

#  ifdef LINUX_SVGALIB
      return (Display *)1;               /* who cares what it is, just not zero :-) */
#  endif

   if(config.mitshm && !is_local_server(dispname)){
      fputs("Disabling MIT-SHM on remote X server\n",stderr);
      config.mitshm=0;
   }
   return display;
}

#ifndef LINUX_SVGALIB

#ifdef MITSHM
static XImage *shm_create(pshminfo,width,height)/* not on Linux console */
XShmSegmentInfo *pshminfo;
int width,height;
{
   XImage *answer;
   answer=XShmCreateImage(display,DefaultVisual(display,screen),
          DefaultDepth(display,screen),ZPixmap,NULL,pshminfo,
          width,height);
   if(!answer){
      fputs("Couldn't create X image\n",stderr);
      return 0;
   }
   pshminfo->shmid=shmget(IPC_PRIVATE,
             answer->bytes_per_line*answer->height,IPC_CREAT|0777);
   if(pshminfo->shmid == -1){
      perror("Couldn't perform shmget");
      return 0;
   }
   pshminfo->shmaddr=answer->data=shmat(pshminfo->shmid,0,0);
   if(!pshminfo->shmaddr){
      perror("Couldn't perform shmat");
      return 0;
   }
   pshminfo->readOnly=0;
   if(!XShmAttach(display,pshminfo)){
      perror("Couldn't perform XShmAttach");
      return 0;
   }
   return answer;
}
#endif
static int image_init()    /* not on Linux console */
{
#ifdef MITSHM
   if(config.mitshm){
      if(!(ximage=shm_create(&xshminfo,hsize,vsize)))return 1;
#ifdef FANCYBORDER
      if(!(bordximage=shm_create(&bordxshminfo,
           hsize+BORDER_WIDTH*2,vsize+BORDER_WIDTH*2)))
         return 1;
#endif
      refresh_frames=config.rrshm;
   } else
#endif
   {
      ximage=XCreateImage(display,DefaultVisual(display,screen),
             DefaultDepth(display,screen),ZPixmap,0,NULL,hsize,vsize,
             8,0);
      if(!ximage){
         perror("XCreateImage failed");
         return 1;
      }
      ximage->data=malloc(ximage->bytes_per_line*ximage->height);
      if(!ximage->data){
         perror("Couldn't get memory for XImage data");
         return 1;
      }
#ifdef FANCYBORDER
      bordximage=XCreateImage(display,DefaultVisual(display,screen),
             DefaultDepth(display,screen),ZPixmap,0,NULL,
             hsize+BORDER_WIDTH*2,vsize+BORDER_WIDTH*2,8,0);
      if(!bordximage){
         perror("XCreateImage failed");
         return 1;
      }
      bordximage->data=malloc(bordximage->bytes_per_line*bordximage->height);
      if(!bordximage->data){
         perror("Couldn't get memory for XImage data");
         return 1;
      }
#endif
      refresh_frames=config.rrnoshm;
   }
   linelen=ximage->bytes_per_line/SCALE;
   if(linelen!=32 && linelen!=256)
      fprintf(stderr,"Line length=%d; expect strange results!\n",linelen);
   if(linelen==32){
      if(BitmapBitOrder(display)==MSBFirst && ImageByteOrder(display)==MSBFirst)
         /*nothing*/;
      else if(BitmapBitOrder(display)==LSBFirst && ImageByteOrder(display)==LSBFirst)
         do_reverse=1;
      else fprintf(stderr,"BitmapBitOrder=%s and ImageByteOrder=%s.\n",
           BitmapBitOrder(display)==MSBFirst?"MSBFirst":"LSBFirst",
           ImageByteOrder(display)==MSBFirst?"MSBFirst":"LSBFirst"),
      fputs("If the display is mixed up, please mail me these values\n",stderr),
      fputs("and describe the display as accurately as possible.\n",stderr);
   }
   image=ximage->data;
#ifdef FANCYBORDER
   bordimage=bordximage->data;
#endif
   return 0;
}

static void notify(argc,argv)   /* not on Linux console */
int *argc;
char **argv;
{
   Pixmap icon;
   XWMHints xwmh;
   XSizeHints xsh;
   XClassHint xch;
   XTextProperty appname, iconname;
   char *apptext;
   char *icontext="xz80";

   icon=XCreatePixmapFromBitmapData(display,root,xz80_bits,
        xz80_width,xz80_height,black,white,DefaultDepth(display,screen));
   apptext="xz80";
   xsh.flags=PSize|PMinSize|PMaxSize;
   xsh.min_width=hsize;
   xsh.min_height=vsize;
   xsh.max_width=hsize+BORDER_WIDTH*2;
   xsh.max_height=vsize+BORDER_WIDTH*2;
   if(!XStringListToTextProperty(&apptext,1,&appname)){
      fputs("Can't create a TextProperty!",stderr);
      return;
   }
   if(!XStringListToTextProperty(&icontext,1,&iconname)){
      fputs("Can't create a TextProperty!",stderr);
      return;
   }
   xwmh.initial_state=NormalState;
   xwmh.input=1;
   xwmh.icon_pixmap=icon;
   xwmh.flags=StateHint|IconPixmapHint|InputHint;
   xch.res_name="xz80";
   xch.res_class="Xz80";
   XSetWMProperties(display,borderwin,&appname,&iconname,argv,
      *argc,&xsh,&xwmh,&xch);
}

#endif

static int get_colours(coltab)
unsigned long coltab[];
{
   XColor xc;
   int i;
   if(DefaultDepthOfScreen(scrptr)==1)config.mono=1;
   if(config.mono)
      for(i=0;i<8;i++)
         coltab[i]=coltab[i+8]=i<4 ? black : white;
   else{
      xc.flags=DoRed|DoGreen|DoBlue;
      for(i=0;i<16;i++){
         xc.red   = rgbvals[i][0];
         xc.green = rgbvals[i][1];
         xc.blue  = rgbvals[i][2];
         if (!XAllocColor(display,cmap,&xc)){
            fprintf(stderr,"XAllocColor failed on colour %d\n",i);
            return 1;
         }
         coltab[i]=xc.pixel;
      }
   }
   return 0;
}
static void reverse_init(){
   int j,l,m;
   unsigned char revval;
   for(l=0;l<256;reverse[l++]=revval)
      if(do_reverse)
         for(m=l,revval=j=0;j<8;j++) {
            revval=(revval<<1)|(m&1);
            m>>=1;
         }
      else revval=l;
}
#if SCALE>1
static void scaleup_init(){
   int j, k, l, m;
   unsigned long bigval; /* SCALING must be <= 4! */
   for(l=0;l<256;scaleup[l++]=bigval){
      for(m=l,bigval=j=0;j<8;j++) {
         for(k=0;k<SCALE;k++)
            bigval=(bigval>>1)|((m&1)<<31);
         m>>=1;
      }
      if(do_reverse) /* hack the bigval if the bytes need to be reversed... */
         bigval=reverse[bigval&255] | (reverse[(bigval>>8)&255]<<8) |
                (reverse[(bigval>>16)&255]<<16) | (reverse[(bigval>>24)&255]<<24);
   }
   /* reverse the shading patterns if white==1 */
   if(white)
      for(j=0;j<16;j++)for(k=0;k<4;k++)
         halftone[j][k]^=255;
}
#endif

void startup(argc,argv)
int *argc;
char **argv;
{
   display=open_display(argc,argv);
   if(!display){
      fputs("Failed to open X display\n",stderr);
      exit(1);
   }
   invert=0;
#  ifdef LINUX_SVGALIB
      get_colours(pixels);
      refresh_frames=config.rrshm;
      linelen=256;
#  else
      screen=DefaultScreen(display);
      scrptr=DefaultScreenOfDisplay(display);
      root=DefaultRootWindow(display);
      white=WhitePixel(display,screen);
      black=BlackPixel(display,screen);
      maingc=XCreateGC(display,root,0,NULL);
      XCopyGC(display,DefaultGC(display,screen),~0,maingc);
      XSetGraphicsExposures(display,maingc,0);
      fggc=XCreateGC(display,root,0,NULL);
      XCopyGC(display,DefaultGC(display,screen),~0,fggc);
      XSetGraphicsExposures(display,fggc,0);
      bggc=XCreateGC(display,root,0,NULL);
      XCopyGC(display,DefaultGC(display,screen),~0,bggc);
      XSetGraphicsExposures(display,bggc,0);
      if(config.private)
         cmap=XCreateColormap(display,root,DefaultVisual(display,screen),
                              AllocNone);
      else cmap=DefaultColormap(display,screen);
      if(get_colours(pixels)){
         if(!config.private){
            fputs("Colour allocation failed - using private cmap\n",stderr);
            config.private=1;
            cmap=XCreateColormap(display,root,DefaultVisual(display,screen),
                                 AllocNone);
            if(get_colours(pixels)){
               fputs("Private colour allocation failed.\n",stderr);
               exit(1);
            }
         } else {
            fputs("Private colour allocation failed.\n",stderr);
            exit(1);
         }
      }
      if(image_init()){
         if(config.mitshm){
            fputs("Failed to create X image - trying again with mitshm off\n",stderr);
            config.mitshm=0;
            if(image_init())exit(1);
         }
         else exit(1);
      }
      reverse_init();
#  endif
#  if SCALE>1
      if(linelen==32)scaleup_init();
#  endif
#ifdef DEV_AUDIO
   if(config.sound)
      if((soundfd=open(config.audio,O_WRONLY|O_CREAT|O_TRUNC,0644))<0){
         config.sound=0;
         perror(config.audio);
         fputs("Error opening audio device for output\n",stderr);
      }
      else {
#ifdef DEV_AUDIO_IN
         if(soundinfd>=0)close(soundinfd);
         if((soundinfd=open(config.audioin,O_RDONLY))<0){
            perror(config.audioin);
            fputs("Error opening audio device for input\n",stderr);
         }
#endif
#ifdef linux
         /* on Linux, try setting the DSP speed correctly as variable
          * sample-rate devices are extremely common */
         int tmp=config.sample; /* use tmp var as this can be varied a bit */
         ioctl(soundfd,SNDCTL_DSP_SPEED,&tmp);
#ifdef LINUX_FRAG
         tmp=0x2000A; /* 2x1k fragments */
         ioctl(soundfd,SNDCTL_DSP_SETFRAGMENT,&tmp);
#endif
#endif
      }
#endif
#ifndef LINUX_SVGALIB
   borderwin=XCreateSimpleWindow(display,root,0,0,
             hsize+BORDER_WIDTH*2,vsize+BORDER_WIDTH*2,0,0,0);
   mainwin=XCreateSimpleWindow(display,borderwin,BORDER_WIDTH,
             BORDER_WIDTH,hsize,vsize,0,0,0);
   notify(argc,argv);
   if(config.private)XSetWindowColormap(display,borderwin,cmap);
   XSelectInput(display,borderwin,KeyPressMask|KeyReleaseMask|
      ExposureMask|EnterWindowMask|LeaveWindowMask|
      StructureNotifyMask);
   XMapRaised(display,borderwin);
   XMapRaised(display,mainwin);
   XSetWindowBackground(display,borderwin,pixels[bordercr]);
   XClearWindow(display,borderwin);
   XFlush(display);
#endif
   refresh_screen=1;
}

#define symshift (symbolshift++,keyports[7]&=0xfd)
#define capshift (capsshift++,keyports[0]&=0xfe)
#define dosym (keyports[7]&=0xfd,keyports[0]|=1)

static int process_keypress(kev)
XKeyEvent *kev;
{
   char buf[3];
   KeySym ks;
   int i;

#  ifdef LINUX_SVGALIB
      ks=kev->key;
#  else
      XLookupString(kev,buf,2,&ks,NULL);
      for(i=0;i<5;i++)if(ks==kemptable[i])kemport|=(1<<i);
#  endif

   switch(ks){
      case XK_F1: return Z80_reset;
      case XK_F3:
         requester("Name of snapshot to save",snapsavename,MAXPATHLEN);
         if(snapsavename[0])return Z80_save;
         else startwatch(1);
         return 0;
      case XK_F4:
         requester("Name of snapshot to load",snapname,MAXPATHLEN);
         if(snapname[0])return Z80_load;
         else startwatch(1);
         return 0;
#ifdef DEV_AUDIO
      case XK_F5:
         if(soundfd>=0){
#ifdef SUN_AUDIO
            itimeron();
#endif
            close(soundfd);
            soundfd=-1;
#ifdef DEV_AUDIO_IN
            close(soundinfd);
            soundinfd=-1;
#endif
         }
         else if((soundfd=open(config.audio,O_WRONLY|O_CREAT|O_TRUNC,0644))<0){
            perror(config.audio);
            fputs("Error opening audio device for output\n",stderr);
         }
         else{
#ifdef SUN_AUDIO
            itimeroff();
#endif
#ifdef DEV_AUDIO_IN
            if((soundinfd=open(config.audioin,O_RDONLY))<0){
               perror(config.audioin);
               fputs("Error opening audio device for input\n",stderr);
            }
#endif
#ifdef linux                  /* as before :-) */
            int tmp=config.sample;
            ioctl(soundfd,SNDCTL_DSP_SPEED,&tmp);
#ifdef LINUX_FRAG
            tmp=0x2000A;      /* 2x1k fragments */
            ioctl(soundfd,SNDCTL_DSP_SETFRAGMENT,&tmp);
#endif
#endif
         }
         return 0;
#endif
      case XK_F6:
         requester("Directory",snapsavename,MAXPATHLEN);
         if(snapsavename[0] && chdir(snapsavename))perror(snapsavename);
         startwatch(1);
         return 0;
      case XK_F7: 
         config.border = (config.border+1)%
#ifdef FANCYBORDER
                                           3;
#else
                                           2;
#endif
#if defined(FANCYBORDER) && SCALE>1
         if(config.border && linelen==32){
            for(i=0;i<vsize+2*BORDER_WIDTH;i++)
               memset(bordimage+i*((hsize+2*BORDER_WIDTH)/8),
                      halftone[bordercr][i%4],(hsize+2*BORDER_WIDTH)/8);
            drawborder();
         } else
#endif
         if(config.border){
            XSetWindowBackground(display,borderwin,pixels[bordercr]);
            XClearWindow(display,borderwin);
         }
         return 0;
      case XK_F8: return Z80_NMI;
      case XK_F9: if(zxpfile)zxpclose();
         else zxpopen();
         return 0;
      case XK_F10: return Z80_quit;
#ifdef DEBUG
      case XK_F11: return Z80_log;
#endif
      case XK_F12:
         if(tapefp)fclose(tapefp),tapefp=0;
         requester("Insert tape",snapsavename,MAXPATHLEN);
         if(!snapsavename[0] && config.tap[0])
            fputs("Tape ejected.\n",stderr);
         if(snapsavename[0]=='<' && !snapsavename[1])
            if(config.tap[0])
               fprintf(stderr,"Tape %s rewound.\n",config.tap);
            else fputs("No tape to rewind\n",stderr);
         else strcpy(config.tap,snapsavename);
         break;
      case XK_Return: 
              keyports[6]&=0xfe; break;
      case XK_Control_L:
              symshift;capshift;break;
      case XK_Shift_L:
              capshift; break;
      case XK_Alt_L: case XK_Alt_R:
      case XK_Meta_L: case XK_Meta_R:
              symshift; break;
      case XK_BackSpace: case XK_Delete:
              capshift; keyports[4]&=0xfe; break;
      case XK_Escape:
              capshift; keyports[7]&=0xfe; break;
      case XK_Tab:
              unshift=!unshift; break;
      case XK_Up:
              unshift||capshift;
              keyports[4]&=0xf7; break;
      case XK_Down:
              unshift||capshift;
              keyports[4]&=0xef; break;
      case XK_Left:
              unshift||capshift;
              keyports[3]&=0xef; break;
      case XK_Right:
              unshift||capshift;
              keyports[4]&=0xfb; break;
      case XK_minus:
              dosym; keyports[6]&=0xf7; break;
      case XK_underscore:
              dosym; keyports[4]&=0xfe; break;
      case XK_equal:
              dosym; keyports[6]&=0xfd; break;
      case XK_plus:
              dosym; keyports[6]&=0xfb; break;
      case XK_semicolon:
              dosym; keyports[5]&=0xfd; break;
      case XK_colon:
              dosym; keyports[0]&=0xfd; break;
      case XK_apostrophe:
              dosym; keyports[4]&=0xf7; break;
#ifndef LINUX_SVGALIB
      case XK_quotedbl:
              dosym; keyports[5]&=0xfe; break;
      case XK_exclam:
              capsshift||dosym; keyports[3]&=0xfe; break;
      case XK_at:
              capsshift||dosym; keyports[3]&=0xfd; break;
      case XK_numbersign:
              capsshift||dosym; keyports[3]&=0xfb; break;
      case XK_dollar:
              capsshift||dosym; keyports[3]&=0xf7; break;
      case XK_percent:
              capsshift||dosym; keyports[3]&=0xef; break;
      case XK_asciicircum:
              if(capsshift)keyports[4]&=0xef;
              else dosym,keyports[6]&=0xef;
              break;
      case XK_ampersand:
              if(capsshift)keyports[4]&=0xf7;
              else dosym,keyports[4]&=0xef;
              break;
      case XK_asterisk:
              if(capsshift)keyports[4]&=0xfb;
              else dosym, keyports[7]&=0xef;
              break;
      case XK_parenleft:
              if(capsshift)keyports[4]&=0xfd;
              else dosym, keyports[4]&=0xfb;
              break;
      case XK_parenright:
              if(capsshift)keyports[4]&=0xfe;
              else dosym, keyports[4]&=0xfd;
              break;
      case XK_comma:
              dosym; keyports[7]&=0xf7; break;
      case XK_less:
              dosym; keyports[2]&=0xf7; break;
      case XK_period:
              dosym; keyports[7]&=0xfb; break;
      case XK_greater:
              dosym; keyports[2]&=0xef; break;
      case XK_slash:
              dosym; keyports[0]&=0xef; break;
      case XK_question:
              dosym; keyports[0]&=0xf7; break;
      case XK_backslash: case XK_bar:
              keyports[0]=keyports[1]=keyports[2]=keyports[3]=
              keyports[4]=keyports[5]=keyports[6]=keyports[7]=0xff;
              unshift=symbolshift=capsshift=kemport=0;
              break;
#endif /* LINUX_SVGALIB */
      default:
#ifdef LINUX_SVGALIB
              ks=(keymap_trans(ks)&255);
#endif
              if(ks>='A' && ks<='Z')keyports[0]&=0xfe;
              if((int)(ks-=32)>=0 && ks<128)
                 keyports[keytable[ks].port]&=keytable[ks].mask;
   }
   return 0;
}

#undef dosym
#undef symshift
#undef capshift
#define dosym (symbolshift || (keyports[7]|=2))
#define symshift (--symbolshift || (keyports[7]|=2))
#define capshift (--capsshift || (keyports[0]|=1))

static void process_keyrelease(kev)
XKeyEvent *kev;
{
   char buf[3];
   KeySym ks;
   int i;

#  ifdef LINUX_SVGALIB
      ks=kev->key;
#  else
      XLookupString(kev,buf,2,&ks,NULL);
      for(i=0;i<5;i++)if(ks==kemptable[i])kemport&=~(1<<i);
#  endif

   switch(ks){
      case XK_Return: 
              keyports[6]|=1; break;
      case XK_Control_L:
              symshift;capshift;break;
      case XK_Shift_L:
              capshift; break;
      case XK_Alt_L: case XK_Alt_R:
      case XK_Meta_L: case XK_Meta_R:
              symshift; break;
      case XK_BackSpace: case XK_Delete:
              capshift; keyports[4]|=1; break;
      case XK_Escape:
              capshift; keyports[7]|=1; break;
      case XK_Tab: break;
      case XK_Up:
              unshift||capshift;
              keyports[4]|=8; break;
      case XK_Down:
              unshift||capshift;
              keyports[4]|=16; break;
      case XK_Left:
              unshift||capshift;
              keyports[3]|=16; break;
      case XK_Right:
              unshift||capshift;
              keyports[4]|=4; break;
      case XK_minus: case XK_underscore:
              dosym; keyports[6]|=8; keyports[4]|=1; break;
      case XK_equal: case XK_plus:
              dosym; keyports[6]|=2; keyports[6]|=4; break;
      case XK_semicolon: case XK_colon:
              dosym; keyports[5]|=2; keyports[0]|=2; break;
#ifndef LINUX_SVGALIB
      case XK_apostrophe: case XK_quotedbl:
              dosym; keyports[4]|=8; keyports[5]|=1; break;
      case XK_exclam: case '1':
              dosym; keyports[3]|=1; break;
      case XK_at: case '2':
              dosym; keyports[3]|=2; break;
      case XK_numbersign: case '3':
              dosym; keyports[3]|=4; break;
      case XK_dollar: case '4':
              dosym; keyports[3]|=8; break;
      case XK_percent: case '5':
              dosym; keyports[3]|=16; break;
      case XK_asciicircum: case '6':
              dosym; keyports[4]|=16; keyports[6]|=16; break;
      case XK_ampersand: case '7':
              dosym; keyports[4]|=8; keyports[4]|=16; break;
      case XK_asterisk: case '8':
              dosym; keyports[4]|=4; keyports[7]|=16; break;
      case XK_parenleft: case '9':
              dosym; keyports[4]|=2; keyports[4]|=4; break;
      case XK_parenright: case '0':
              dosym; keyports[4]|=1; keyports[4]|=2; break;
      case XK_comma: case XK_less:
              dosym; keyports[7]|=8; keyports[2]|=8; break;
      case XK_period: case XK_greater:
              dosym; keyports[7]|=4; keyports[2]|=16; break;
      case XK_slash: case XK_question:
              dosym; keyports[0]|=16; keyports[0]|=8; break;
      case XK_backslash: case XK_bar: break;
#endif /* LINUX_SVGALIB */
      default:
#ifdef LINUX_SVGALIB
              ks=(keymap_trans(ks)&255);
#endif
              if(!capsshift)keyports[0]|=1;
              if((int)(ks-=32)>=0 && ks<96)
                 keyports[keytable[ks].port]|=~keytable[ks].mask;
   }
   return;
}

int check_events(flag)
int flag;
{
   static int answer=0;
   int i;
   static XEvent xev;
#  ifndef LINUX_SVGALIB
      XConfigureEvent *conf_ev;
      XCrossingEvent *cev;
#  endif
   if(load_snap)answer=4,load_snap=0;
#ifndef LINUX_SVGALIB
   while (XEventsQueued(display,QueuedAfterReading)){
      XNextEvent(display,&xev);
#else
   linux_get_keys(&xev);
   while(xev.key!=NoKeyPressed){
#endif
      switch(xev.type){
#ifndef LINUX_SVGALIB
         case Expose: refresh_screen=1;
            break;
         case ConfigureNotify:
            conf_ev=(XConfigureEvent *)&xev;
            XMoveWindow(display,mainwin,(conf_ev->width-hsize)/2,
                        (conf_ev->height-vsize)/2);
            break;
         case MapNotify: case UnmapNotify: case ReparentNotify: case NoExpose:
            break;
         case EnterNotify:
            cev=(XCrossingEvent *)&xev;
            if(cev->detail!=NotifyInferior)
               XAutoRepeatOff(display),XFlush(display);
            break;
         case LeaveNotify:
            cev=(XCrossingEvent *)&xev;
            if(cev->detail!=NotifyInferior)
               XAutoRepeatOn(display),XFlush(display);
            break;
#endif
         case KeyPress: i=process_keypress((XKeyEvent *)&xev);
            if(!answer)answer=i;
            break;
         case KeyRelease: process_keyrelease((XKeyEvent *)&xev);
            break;
         default:
            fprintf(stderr,"unhandled X event, type %d\n",xev.type);
      }
#     ifdef LINUX_SVGALIB
         linux_get_keys(&xev);      /* ready for next time around the loop */
#     endif
   }
   if(flag){
      i=answer;
      answer=0;
      return i;
   }
   else return 0;
}

#ifdef FANCYBORDER
void drawborder(){  /* put border from ximage on to the screen */
#ifdef MITSHM
   if(config.mitshm){
      XShmPutImage(display,borderwin,maingc,bordximage,0,0,0,0,
                   hsize+2*BORDER_WIDTH,BORDER_WIDTH,0);
      XShmPutImage(display,borderwin,maingc,bordximage,
                   0,BORDER_WIDTH,0,BORDER_WIDTH,BORDER_WIDTH,vsize,0);
      XShmPutImage(display,borderwin,maingc,bordximage,
                   hsize+BORDER_WIDTH,BORDER_WIDTH,hsize+BORDER_WIDTH,
                   BORDER_WIDTH,BORDER_WIDTH,vsize,0);
      XShmPutImage(display,borderwin,maingc,bordximage,
                   0,vsize+BORDER_WIDTH,0,vsize+BORDER_WIDTH,
                   hsize+2*BORDER_WIDTH,BORDER_WIDTH,0);
   } else
#endif
   {
      XPutImage(display,borderwin,maingc,bordximage,0,0,0,0,
                   hsize+2*BORDER_WIDTH,BORDER_WIDTH);
      XPutImage(display,borderwin,maingc,bordximage,
                   0,BORDER_WIDTH,0,BORDER_WIDTH,BORDER_WIDTH,vsize);
      XPutImage(display,borderwin,maingc,bordximage,
                   hsize+BORDER_WIDTH,BORDER_WIDTH,hsize+BORDER_WIDTH,
                   BORDER_WIDTH,BORDER_WIDTH,vsize);
      XPutImage(display,borderwin,maingc,bordximage,
                   0,vsize+BORDER_WIDTH,0,vsize+BORDER_WIDTH,
                   hsize+2*BORDER_WIDTH,BORDER_WIDTH);
   }
}
static int oldadd=0,oldx=0,oldy=0;
void fillborder(time) /* fill the border with the previous colour */
int time;             /* up until the specified time */
{
   int tim=time-frame_sync+(BORDER_EPIX*linecycles+BORDER_EPIX/2);
   int newy=tim/linecycles;
   int newx=(tim%linecycles)/4;
   int add,cy,lx;
   int pix=pixels[bordercr];
   if(newy<0 || newx<0)return;
   if(newx>SCREEN_EPIX/8)newx=SCREEN_EPIX/8;
   if(newy>=192+2*BORDER_EPIX)newy=192+2*BORDER_EPIX, newx=0;
#if SCALE==1
   if(linelen==32)add=newy*(SCREEN_EPIX/8)+newx,pix=pix?255:0;
   else add=newy*SCREEN_EPIX+8*newx;
   if(add>oldadd)memset(bordimage+oldadd,pix,add-oldadd);
   oldadd=add;
#else
   if(newy==oldy)lx=(newx-oldx)*SCALE; 
   else lx=(SCREEN_EPIX/8-oldx)*SCALE;
   if(linelen==32){
      pix=bordercr;
      add=(oldy*(SCREEN_EPIX/8)*SCALE+oldx)*SCALE;
      for(cy=oldy*SCALE;cy<(newy+1)*SCALE;cy++){
         if(cy==(oldy+1)*SCALE)lx=SCREEN_WIDTH/8,add-=oldx*SCALE;
         if(cy==newy*SCALE && oldy<newy)lx=newx*SCALE;
         if(lx)memset(bordimage+add,halftone[pix][cy%4],lx);
         add+=SCREEN_WIDTH/8;
      }
   } else {
      lx*=8;
      add=(oldy*SCREEN_EPIX*SCALE+oldx*8)*SCALE;
      for(cy=oldy*SCALE;cy<(newy+1)*SCALE;cy++){
         if(cy==(oldy+1)*SCALE)lx=SCREEN_WIDTH,add-=oldx*8*SCALE;
         if(cy==newy*SCALE && oldy<newy)lx=newx*8*SCALE;
         if(lx)memset(bordimage+add,pix,lx);
         add+=SCREEN_WIDTH;
      }
   }
   oldy=newy; oldx=newx;
#endif
}
#endif
void refresh(){
   int h,i,j,k,l,m,n;
   int attr;
   int mono=config.mono;
   int flashing=config.flashing;
#ifdef MITSHM
   int mitshm=config.mitshm;
#endif
   int xmin,xmax,ymin,ymax;
   unsigned char *square;
   unsigned char *dest;
   unsigned char val;
#if FANCYBORDER
#if SCALE>1
   if(!borderchange && refresh_screen && config.border==2)drawborder();
#endif
   if(borderchange && config.border==2){
      if(--borderchange){ /* Border is stripey */
         fillborder(config.cycles);
         drawborder();
      } else {            /* Border is all one colour */
#if SCALE>1
         if(linelen==32){ /* the border should be halftoned */
            for(i=0;i<vsize+2*BORDER_WIDTH;i++)
               memset(bordimage+i*((hsize+2*BORDER_WIDTH)/8),
                      halftone[bordercr][i%4],(hsize+2*BORDER_WIDTH)/8);
            drawborder();
         } else
#endif
         {
            XSetWindowBackground(display,borderwin,pixels[bordercr]);
            XClearWindow(display,borderwin);
         }
      }
      if(!screen_dirty && !refresh_screen)XFlush(display);
      oldadd=oldx=oldy=0;
   } else
#endif
   {
      if(borderchange>1)XClearWindow(display,borderwin),XFlush(display);
      borderchange=0;
   }
   if(!screen_dirty)
      if(refresh_screen){
#ifdef MITSHM
         if(mitshm)
            XShmPutImage(display,mainwin,maingc,ximage,0,0,0,0,hsize,vsize,0);
         else
#endif
            XPutImage(display,mainwin,maingc,ximage,0,0,0,0,hsize,vsize);
         XFlush(display);
         refresh_screen=0;
         return;
      }
      else return;
#if defined(SunKludge) && SCALE>1
   if(linelen==32)  /* This makes XSunMono less likely to crash... */
      xmin=ymin=0,xmax=31,ymax=23;
   else
#endif
#ifdef MITSHM
   if(refresh_screen && mitshm)
      xmin=ymin=0,xmax=31,ymax=23;
   else
#endif
   xmin=31,ymin=23,xmax=ymax=0;
   for(j=k=0;j<24;j++)
      for(i=0;i<32;i++,k++)
         if(squares[k]){
            attr=attrfile[k];
            square=dfile+((k&0xff)|((k>>8)<<11));
            if(linelen==256){
               int paper=(attr>>3)&15;
               int ink=(attr&7)|((attr&64)>>3);
               dest=image+((i<<3)|((j*SCALE)<<11))*SCALE;
               if(ink==paper){
                  for(l=0;l<8*SCALE;l++,dest+=248*SCALE)
                     for(m=0;m<8*SCALE;m++,dest++)
                        dest[0]=pixels[ink];
               } else {
                  if(mono)
                     if(ink>paper)ink=7,paper=0;
                     else ink=0,paper=7;
                  if(invert && (attr&128)){int t=ink;ink=paper;paper=t;}
                  else if(!flashing && (attr&128)) paper|=8,ink|=8;
                  for(l=0;l<8;l++,square+=256)
#if SCALE>1
                     for(n=0;n<SCALE;n++)
#endif
                  {
                     val=square[0];
                     for(m=0;m<8;m++,val<<=1)
#if SCALE>1
                        for(h=0;h<SCALE;h++)
#endif
                           dest++[0]=pixels[(val&128)?ink:paper];
                     dest+=248*SCALE;
                  }
               }
            }
            else{
#if SCALE==1
               dest=image+(i|(j<<8));
               if((attr&7)==((attr>>3)&7)){
                  attr=(((attr&7)>4)^white)?0:255;
                  for(l=0;l<8;l++)dest[0]=attr,dest+=32;
               } else {
                  attr=((attr&7)>((attr>>3)&7))^(black==1)^
                        (invert&&(attr&0x80))?0:255;
                  for(l=0;l<8;l++)
                     dest[0]=reverse[square[0]^attr],dest+=32,square+=256;
               }
#else
               int paper=(attr>>3)&15;
               int ink=(attr&7)|((attr&64)>>3);
               unsigned char pmask,ipmask;
               int shaded=config.shaded;
               unsigned long bigval;
               dest=image+(i|((j*SCALE)<<8))*SCALE;
               if(invert && (attr&0x80)){int t=ink;ink=paper;paper=t;}
               else if(!flashing && (attr&0x80))ink|=8,paper|=8;
               if(!shaded){
                  if(ink==paper)pmask=(ink&7)<4?black*255:white*255,
                                ipmask=0;
                  else if((ink>paper)^white)pmask=ipmask=255;
                  else pmask=0,ipmask=255;
               }
               h=0;
               for(l=0;l<8;l++){
                  bigval=scaleup[square[0]];
                  square+=256;
                  for(m=0;m<SCALE;m++){
                     if(shaded){
                        pmask=halftone[paper][h];
                        ipmask=halftone[ink][h]^pmask;
                        if(++h==4)h=0;
                     }
                     dest[0]=shade(bigval>>24,ipmask,pmask);
                     dest[1]=shade(bigval>>16,ipmask,pmask);
#if SCALE>2
                     dest[2]=shade(bigval>> 8,ipmask,pmask);
#endif
#if SCALE>3
                     dest[3]=shade(bigval    ,ipmask,pmask);
#endif
                     dest+=32*SCALE;
                  }
               }
#endif
            }
#ifdef MITSHM
            if(mitshm){
               if(i>xmax)xmax=i;
               if(i<xmin)xmin=i;
               if(j>ymax)ymax=j;
               if(j<ymin)ymin=j;
            }
            else
#endif
            if(!refresh_screen)
               XPutImage(display,mainwin,maingc,ximage,
                   i*8*SCALE,j*8*SCALE,i*8*SCALE,j*8*SCALE,
                   8*SCALE,8*SCALE);
            squares[k]=0;
         }
#ifdef MITSHM
   if(mitshm)
      XShmPutImage(display,mainwin,maingc,ximage,
         xmin*8*SCALE,ymin*8*SCALE,xmin*8*SCALE,ymin*8*SCALE,
         (xmax-xmin+1)*8*SCALE,(ymax-ymin+1)*8*SCALE,0);
   else
#endif
   if(refresh_screen)
      XPutImage(display,mainwin,maingc,ximage,0,0,0,0,hsize,vsize);
   XFlush(display);
   refresh_screen=screen_dirty=0;
}

void flash(){
   int i;
   invert=!invert;
   for(i=0;i<768;i++)if(attrfile[i]&0x80)squares[i]=screen_dirty=1;
}

void closedown(){
   zxpclose();
#if defined(MITSHM) && !defined(LINUX_SVGALIB)
   if(config.mitshm){
      XShmDetach(display,&xshminfo);
      XDestroyImage(ximage);
      shmdt(xshminfo.shmaddr);
      shmctl(xshminfo.shmid,IPC_RMID,0);
#ifdef FANCYBORDER
      XShmDetach(display,&bordxshminfo);
      XDestroyImage(bordximage);
      shmdt(bordxshminfo.shmaddr);
      shmctl(bordxshminfo.shmid,IPC_RMID,0);
#endif
   }
#endif
   XAutoRepeatOn(display);
   XCloseDisplay(display);
#ifdef DEV_AUDIO
   if(soundfd>=0)close(soundfd);
#ifdef DEV_AUDIO_IN
   if(soundinfd>=0)close(soundinfd);
#endif
#endif
   if(tapefp)fclose(tapefp);
}

static void zxpopen(){
   if(!config.print[0]){
      requester("Name of printer file",config.print,sizeof config.print);
      startwatch(1);
      if(!config.print[0])return;
   }
   if(!(zxpfile=fopen(config.print,"w"))){
      perror(config.print);
      return;
   }
   fputs("#define zxp_width 256\n#define zxp_height           \n",zxpfile);
   fputs("static char zxp_bits[] = {\n",zxpfile);
   if(ferror(zxpfile)){
      perror(config.print);
      fclose(zxpfile);
      zxpfile=0;
      return;
   }
   zxpstylus=zxpspeed=zxpheight=zxpnewspeed=0;
   zxppixel=-1;
   fprintf(stderr,"Print file %s open\n",config.print);
}

static void zxpclose(){
   if(!zxpfile)return;
   out(config.cycles,0,251,4); /* stop the printer */
   /* finish the file with a brace instead of a comma */
   fputs("        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \n",zxpfile);
   fputs("        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \n",zxpfile);
   fputs("        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \n",zxpfile);
   fputs("        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};\n",zxpfile);
   zxpheight++;
   if(fseek(zxpfile,41,0))
      fputs("warning: couldn't seek to write image height\n",stderr);
   else fprintf(zxpfile,"%d",zxpheight);
   fclose(zxpfile);
   zxpfile=0;
   fprintf(stderr,"Printer output written on %s\n",config.print);
}

static void zxpout(){
   int i,j,k,d;
   zxpheight++;
   for(i=0;i<32;i++){
      if(i%8==0)fputs("        ",zxpfile);
      fputs("0x",zxpfile);
      for(k=0;k<2;k++){
         for(d=j=0;j<4;j++)
            d=(d<<1)+(zxpline[i*8+7-k*4-j]?1:0);
         putc("0123456789abcdef"[d],zxpfile);
      }
      fputs(", ",zxpfile);
      if(i%8==7)putc('\n',zxpfile);
   }
}

#ifdef DEV_AUDIO_IN
static signed char audioinbuf[AUDIOBUFLEN];
static int centre=0;
/*static int lastin=0;*/
#endif

/* Input from the I/O port whose high address byte is "hi" and whose
   low address byte is "lo".  The input occurred "time" clock cycles
   after the start of the current frame.  The result from the function
   is the 8-bit unsigned input from the port, plus (t<<8) where t is
   the number of wait states introduced by the I/O device. */
unsigned int in(time,hi,lo)
unsigned long time;
unsigned char hi,lo;
{
   if(lo==254){
      int i;
      unsigned char ans=0xff;
#if defined(DEV_AUDIO_IN) && defined(DEV_AUDIO)
      if(soundinfd>=0){
/*       int trig=config.trigger;   */
/*       int rem=time%bytecycles;   */
         int pos=time/bytecycles;
         int val /* ,last */ ;
         if(pos==framebytes)pos--;
         val=audioinbuf[pos]-centre;
/*       last=pos>0?audioinbuf[pos-1]-centre:lastin;       (this would have done
         val=(val*rem+last*(bytecycles-rem))/bytecycles;    linear interpolation)
         if(val<-trig)ans=0xbf;                            (this would have added
         else if(val<trig && val<rand()%(2*trig)-trig)      a probabilistic mode for
              ans=0xbf;                                     sampling quiet sounds)
*/
         if(val<0)ans=0xbf;
      }
#endif
  /*  check_events(0);   it seems to be OK without this */
      for(i=0;i<8;i++){
         if(!(hi&1))ans&=keyports[i];
         hi>>=1;
      }
      return ans;
   }
#  ifndef LINUX_SVGALIB
      if(lo==31)return kemport;
#  endif
   if(lo==251 && zxpfile){
      if(!zxpspeed)return 0x3e;
      else {
         int frame=frames-zxpframes;
         int cycles=time-zxpcycles;
         int pix=zxppixel;
         int sp=zxpnewspeed;
         int x;
         unsigned char ans;
         int cpp=440/zxpspeed;
         if(frame>400)frame=400;
         cycles+=frame*config.cycles;
         x=cycles/cpp-64;        /* x-coordinate reached */
         while(x>320){           /* if we are on another line, */
            pix=-1;              /* find out where we are */
            x-=384;
            if(sp){
               x=(x+64)*cpp;
               cpp=440/sp;
               x=x/cpp-64;
               sp=0;
            }
         }
         if((x>-10 && x<0) | zxpstylus)ans=0xbe;
         else ans=0x3e;
         if(x>pix)ans|=1;
         return ans;
      }
   }
   return 255;
}

#ifdef DEV_AUDIO
static signed char audiobuf[AUDIOBUFLEN];
static int lastpos=0,lastrem=0,lastval=0;
static int accum=0;
static int sounds=1;


/* This array converts from unsigned character (=signed char plus 128)
   to u-law encoding as used by /dev/audio, so that I don't have to
   include <multimedia/ulaw2linear.h> which may or may not be present */
static unsigned char ulaw[256]={
        0,  0,  0,  0,  1,  1,  1,  1,  2,  2,  2,  2,  3,  3,  3,  3,
        4,  4,  4,  4,  5,  5,  5,  5,  6,  6,  6,  6,  7,  7,  7,  7,
        8,  8,  8,  8,  9,  9,  9,  9, 10, 10, 10, 10, 11, 11, 11, 11,
       12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15,
       16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22, 23, 23,
       24, 24, 25, 25, 26, 26, 27, 27, 28, 28, 29, 29, 30, 30, 31, 31,
       32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
       48, 50, 52, 54, 56, 58, 60, 61, 63, 67, 71, 75, 79, 87, 95,111,
      240,224,216,208,204,200,196,192,190,188,186,184,182,180,178,176,
      175,174,173,172,171,170,169,168,167,166,165,164,163,162,161,160,
      159,159,158,158,157,157,156,156,155,155,154,154,153,153,152,152,
      151,151,150,150,149,149,148,148,147,147,146,146,145,145,144,144,
      143,143,143,143,142,142,142,142,141,141,141,141,140,140,140,140,
      139,139,139,139,138,138,138,138,137,137,137,137,136,136,136,136,
      135,135,135,135,134,134,134,134,133,133,133,133,132,132,132,132,
      131,131,131,131,130,130,130,130,129,129,129,129,128,128,128,128
};
#ifdef DEV_AUDIO_IN
/* This array is the reverse of the above, converting u-law inputs into signed
   chars. */
static signed char walu[256]={
      -126,-122,-118,-114,-110,-106,-102,-98,-94,-90,-86,-82,-78,-74,-70,-66,
      -63,-61,-59,-57,-55,-53,-51,-49,-47,-45,-43,-41,-39,-37,-35,-33,
      -32,-31,-30,-29,-28,-27,-26,-25,-24,-23,-22,-21,-20,-19,-18,-17,
      -16,-15,-15,-14,-14,-13,-13,-12,-12,-11,-11,-10,-10, -9, -9, -8,
       -8, -8, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4,
       -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2,
       -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1,
       -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  0,
      126,121,117,113,109,105,101, 97, 93, 89, 85, 81, 77, 73, 69, 65,
       62, 60, 58, 56, 54, 52, 50, 48, 46, 44, 42, 40, 38, 36, 34, 32,
       31, 30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20, 19, 18, 17, 16,
       15, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10,  9,  9,  8,  8,  7,
        7,  7,  6,  6,  6,  6,  5,  5,  5,  5,  4,  4,  4,  4,  3,  3,
        3,  3,  3,  3,  2,  2,  2,  2,  2,  2,  2,  2,  1,  1,  1,  1,
        1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
};
#endif
#endif

/* Things to do for the hardware at the end of each frame. */
void eachframe(){
#ifdef DEV_AUDIO
   if(soundfd>=0 
#ifndef SUN_AUDIO  /* silence is not output when SUN_AUDIO is off */
       && sounds
#endif
   ){
      int i,j;
      int volume=config.volume;
      audiobuf[lastpos]=(accum+lastval*(bytecycles-lastrem))/bytecycles;
      for(i=lastpos+1;i<framebytes;i++)audiobuf[i]=lastval;
      for(i=0;i<framebytes;i++){
         j=((audiobuf[i]*volume)/256)+128;
         if(j<0)j=0;
         else if(j>255)j=255;
         audiobuf[i]=ulaw[j];
      }
#ifdef SUN_AUDIO
#ifndef linux
      ioctl(soundfd,AUDIO_DRAIN,0);
#else
      ioctl(soundfd,SNDCTL_DSP_POST,0);       /* similar call for Linux */
#endif
#endif
      write(soundfd,audiobuf,framebytes);
   }
   accum=sounds=lastpos=lastrem=0;
#ifdef DEV_AUDIO_IN
   if(soundinfd>=0){
      long cent=centre*9*framebytes;
      int i;
/*    lastin=audioinbuf[framebytes-1]-centre; */
      read(soundinfd,audioinbuf,framebytes);
      for(i=0;i<framebytes;i++){
         cent+=(audioinbuf[i]=walu[(unsigned char)audioinbuf[i]]);
      }
      centre=cent/(framebytes*10);
#endif
   }
#endif
}

/* Output "val" to the I/O port whose high address byte is "hi" and
   whose low address byte is "lo".  The output occurred "time" clock
   cycles after the start of the current frame.  The result from the
   function is the number of wait states introduced by the I/O device.*/
unsigned int out(time,hi,lo,val)
unsigned long time;
unsigned char hi,lo,val;
{
   if(lo==254){
      int newbord=val&7;
      if(newbord!=bordercr){
#ifdef FANCYBORDER
         if(config.border==2){
            if(refresh_ct==1)borderchange=2,fillborder(time);
            else borderchange=1;
         } else
#endif
         if(config.border){
            XSetWindowBackground(display,borderwin,pixels[newbord]);
            /* clear takes a long time, so only do once per refresh */
            if(!borderchange++)XClearWindow(display,borderwin);
         }
         bordercr=newbord;
      }
#ifdef DEV_AUDIO
      if(soundfd>=0)
      {
         int value=(val&24)*8-96;
         if(value!=lastval){
            int pos=time/bytecycles;
            int rem=time%bytecycles;
            if(pos==lastpos)accum+=lastval*(rem-lastrem);
            else{
               int i;
               audiobuf[lastpos]=(accum+lastval*(bytecycles-lastrem))/bytecycles;
               accum=lastval*rem;
               for(i=lastpos+1;i<pos;i++)audiobuf[i]=lastval;
               lastpos=pos;
            }
            lastrem=rem;
            lastval=value;
            sounds=1;
         }
      }
#endif
      return 0;
   }
   if(lo==251 && zxpfile){
      if(zxpspeed){
         int frame=frames-zxpframes;
         int cycles=time-zxpcycles;
         int i,x;
         int cpp=440/zxpspeed;
         if(frame>400)frame=400; /* limit height of blank paper */
         cycles+=frame*config.cycles;
         x=cycles/cpp-64;        /* x-coordinate reached */
         for(i=zxppixel;i<x && i<256;i++)zxpline[i]=zxpstylus;
         if(x>=256 && zxppixel<256)zxpout();
         while(x>=320){          /* move to next line */
            zxpcycles+=cpp*384;
            if(zxpcycles>=config.cycles)
               zxpcycles-=config.cycles,zxpframes++;
            x-=384;
            if(zxpnewspeed){
               zxpspeed=zxpnewspeed;
               zxpnewspeed=0;
               x=(x+64)*cpp;
               cpp=440/zxpspeed;
               x=x/cpp-64;
            }
            for(i=0;i<x && i<256;i++)zxpline[i]=zxpstylus;
            if(x>=256)zxpout();
         }
         if(x<0)x=-1;
         if(val&4){
            if(x>=0 && x<256){
               for(i=x;i<256;i++)zxpline[i]=zxpstylus;
               zxpout();
            }
            zxpspeed=zxpstylus=0;
         } else {
            zxppixel=x;
            zxpstylus=val&128;
            if(x<0)zxpspeed=(val&2)?1:2;
            else {
               zxpnewspeed=(val&2)?1:2;
               if(zxpnewspeed==zxpspeed)zxpnewspeed=0;
            }
         }
      } 
      else if(!(val&4)){
         zxpspeed=(val&2)?1:2;
         zxpframes=frames;
         zxpcycles=time;
         zxpstylus=val&128;
         zxppixel=0;
      }
      return 0;
   }
   return 0;
}
