/* Miscellaneous definitions for xz80, copyright (C) 1994 Ian Collier.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define VERSION "0.1d"

#include<sys/param.h>
#define flash_speed 25
#define BORDER_EPIX   32      /* number of emulated pixels in border */
#define SCREEN_EPIX  (256+2*BORDER_EPIX)
#define BORDER_WIDTH (BORDER_EPIX*SCALE)
#define SCREEN_WIDTH (SCREEN_EPIX*SCALE)
#define K48_SYNC  14116       /* timing information for the border */
#define K128_SYNC 14368
#define K48_LINE  224
#define K128_LINE 228
#define K48_CYCLES  69888     /* estimated clock cycles per frame */
#define K128_CYCLES 70908
#ifndef LIBPATH
#define LIBPATH ".:/usr/local/xz80/lib"
#endif
#ifndef AUDIODEV
#define AUDIODEV "/dev/audio"
#endif

#define KEYSYMLEN 40

#define XZX_MAGIC "ZXF1"

#define Z80_quit  1
#define Z80_NMI   2
#define Z80_reset 3
#define Z80_load  4
#define Z80_save  5
#define Z80_log   6

extern unsigned char *memptr[];
extern int memattr[];
extern char squares[];
extern unsigned char *dfile;
extern unsigned char *attrfile;
extern int screen_dirty;
extern int refresh_screen;
extern int refresh_frames;
extern int refresh_ct;
extern unsigned long frames;
extern unsigned char snapa,snapf,snapb,snapc,snapd,snape,snaph,snapl,
             snapa1,snapf1,snapb1,snapc1,snapd1,snape1,snaph1,snapl1,
             snapi,snapr,snapiff1,snapiff2,snapim;
extern unsigned short snapix,snapiy,snapsp,snappc;
extern int load_snap;
extern char snapname[];
extern char snapsavename[];
extern int hsize,vsize;
extern unsigned char keyports[9];
extern int bordercr;
extern int borderchange;
extern int soundfd;
extern FILE *tapefp;
extern int taperead;
extern struct _kempinit {char *resource;char keysym[KEYSYMLEN];}kempinit[];

extern struct _config {
   int mitshm;
   int mono;
   int private;
   int shaded;
   int flashing;
   int sound;
   int sample;
   int usec;        /* microseconds per frame */
   int cycles;      /* clock cycles per frame */
   int input_hack;
   int rrshm;
   int rrnoshm;
   int volume;
#ifdef DEBUG
   char log[MAXPATHLEN];
#endif
   char rom[MAXPATHLEN];
   char lib[MAXPATHLEN];
#ifdef DEV_AUDIO
   char audio[MAXPATHLEN];
#ifdef DEV_AUDIO_IN
   char audioin[MAXPATHLEN];
#endif
#endif
   char print[MAXPATHLEN];
   char tap[MAXPATHLEN];
   int printer;
   int multiload;
   int save;
   int load;
   int noprompt;
   int border;
} config;

unsigned int in();
unsigned int out();
int interrupt();
int snapload();
void snapsave();
void mainloop();
void eachframe();
void itimeron();
void itimeroff();
void startwatch();
unsigned long stopwatch();
void requester();
int loader();
int saver();
void multiloader();
void usage();
void version();
void drawborder();

#define fetch(x) (memptr[(unsigned short)(x)>>14][(x)&16383])
#define fetch2(x) ((fetch((x)+1)<<8)|fetch(x))

#define store(x,y) do {\
          unsigned short off=(x)&16383;\
          unsigned char page=(unsigned short)(x)>>14;\
          int attr=memattr[page];\
          if(attr){\
             memptr[page][off]=(y);\
             if(--attr && off<0x1b00){\
                screen_dirty=1;\
                if(off>=0x1800)squares[off-0x1800]=1;\
                else squares[((off>>11)<<8)|(off&0xff)]=1;\
       }  }  } while(0)

#define store2b(x,hi,lo) do {\
          unsigned short off=(x)&16383;\
          unsigned char page=(unsigned short)(x)>>14;\
          int attr=memattr[page];\
          if(off==16383){\
             if(attr)memptr[page][off]=(lo);\
             page=(page+1)&3;\
             attr=memattr[page];\
             if(attr){\
                memptr[page][0]=(hi);\
                if(--attr)screen_dirty=1,squares[0]=1;\
          }  } else if(attr) { \
             memptr[page][off]=(lo);\
             memptr[page][off+1]=(hi);\
             if(--attr && off<0x1b00){\
                screen_dirty=1;\
                if(off>=0x1800){int squ;\
                   squares[squ=off-0x1800]=1;\
                   if(++squ<768)squares[squ]=1;\
                }\
                else {\
                   squares[((off>>11)<<8)|(off&0xff)]=1;\
                   if(++off==0x1800)squares[0]=1;\
                   else squares[((off>>11)<<8)|(off&0xff)]=1;\
          }  }  }\
       } while(0)

#define store2(x,y) store2b(x,(y)>>8,y)

#ifdef __GNUC__
static void inline storefunc(unsigned short ad,unsigned char b){
   store(ad,b);
}
#undef store
#define store(x,y) storefunc(x,y)

static void inline store2func(unsigned short ad,unsigned char b1,unsigned char b2){
   store2b(ad,b1,b2);
}
#undef store2b
#define store2b(x,hi,lo) store2func(x,hi,lo)
#endif

#define bc ((b<<8)|c)
#define de ((d<<8)|e)
#define hl ((h<<8)|l)
